// #######################################################
//  THEREMINO TESTER
// #######################################################
//  Based on the work and research of:
//   - Karl-Heinz K¨ubbeler - kh_kuebbeler@web.de
//   - Markus Reschke - madires@theca-tabellaria.de
//   - Theremino Team and many other experimenters
// =======================================================
//  This version works on NANO connected to PC
//  No display, no battery, no buttons are required
//  The only required components are:
//   3 resistors of 680 ohm 
//   3 resistors of 470 kohm
// #######################################################

// =======================================================
//  ARDUINO NANO CONNECTIONS
// =======================================================
// TEST-PROBE 1 connected to Pin A0
// TEST-PROBE 2 connected to Pin A1
// TEST-PROBE 3 connected to Pin A2

// RESISTOR 680 ohm  from Pin A0 to Pin D2
// RESISTOR 470 Kohm from Pin A0 to Pin D3
// RESISTOR 680 ohm  from Pin A1 to Pin D4
// RESISTOR 470 Kohm from Pin A1 to Pin D5
// RESISTOR 680 ohm  from Pin A2 to Pin D6
// RESISTOR 470 Kohm from Pin A2 to Pin D7

// START-TEST PULSE (optional) from Pin A3 to GND
// (internal pull-up - no resistor needed)

// PIEZO or MAGNETIC BUZZER (optional) from Pin D10 to GND
// =======================================================


// ======================================================================================
//   CONFIG OPTIONS
// ======================================================================================

// --------------------------------------------------------------------------------------
// #define START_AS_FAST_TESTER to test continuity with Beep
// #undefine START_AS_FAST_TESTER to start as classic Transistor Tester
// --------------------------------------------------------------------------------------
#define START_AS_FAST_TESTER

// --------------------------------------------------------------------------------------
// #define RESISTORS_ON_D2D7 to use resistors on D2 to D7    (processor PORTD bits 2 to 7)
// #undefine RESISTORS_ON_D2D7 to use resistors on D8 to D13 (processor PORTB bits 0 to 5)
// -------------------------------------------------------------------------------------- 
//  D8 to D13 are the classic Transistor Tester configuration
//  but on the Nano there is a LED connected to D13
//  so to use D8 to D13 you MUST remove the LED resistor
// --------------------------------------------------------------------------------------
#define RESISTORS_ON_D2D7

// --------------------------------------------------------------------------------------
// The standard values are 680 ohm and 470 Kohm
// Please select resistors measuring them with a tester or use precision resistors ( +/- 1 % )
// OK if your resistors are in the range 675 to 685 and 465K to 475K
// otherwise calibrate the following values:  
// --------------------------------------------------------------------------------------           
#define R_L_VAL 6800    // This will be define a 680 ohm  -- Multiplied by 10    
#define R_H_VAL 47000   // This will be define a 470 Kohm -- Divided by 10     

// --------------------------------------------------------------------------------------
// CAPACITORS
// --------------------------------------------------------------------------------------
// REF_C_LOW_KORR corrects the reference voltage for low capacitors (<40uF) and has mV units.
// Positive values will reduce measurement results.
#define REF_C_LOW_KORR 19

// REF_C_HI_KORR defines a correction of 0.1% units for big capacitors measurement.
// Positive values will reduce measurement results.
#define REF_C_HI_KORR 8

// The CAP_EMPTY_LEVEL  defines the empty voltage level for capacitors in mV.
// Choose a higher value, if your Tester reports "Cell!" by unloading capacitors.
#define CAP_EMPTY_LEVEL 4  

// CABLE_CAP defines the capacity (pF) of 12cm cable with clip at the terminal pins.
#define CABLE_CAP 2

// --------------------------------------------------------------------------------------
// INDUCTORS
// --------------------------------------------------------------------------------------
// REF_L_LOW_KORR corrects the reference voltage for little inductors (< 10 mH) measurement.
// Positive values will reduce measurement results.
#define REF_L_LOW_KORR 220

// REF_L_HI_KORR corrects the reference voltage for big inductors measurement.
// Positive values will reduce measurement results.
#define REF_L_HI_KORR 56

// --------------------------------------------------------------------------------------
// RESISTORS
// --------------------------------------------------------------------------------------
// The CAP_ESR_ZERO value define the zero value for capacitors ESR measurement
// Positive values will decrease measurement results (units = 0.01 Ohm).
#define CAP_ESR_ZERO 30

// The RES_ESR_ZERO value defines the zero for the low value resistors (< 100 ohm) 
// Positive values will decrease measurement results (units = 0.01 Ohm).
#define RES_ESR_ZERO 95

// REF_R_KORR corrects the reference voltage for resistor mesurements
// Positive values will decrease measurement results.
#define REF_R_KORR 0

// RH_OFFSET : systematic offset of resistor measurement with RH (470k) 
// Resolution is 1 Ohm, for example 100 defines a offset of +100 Ohm
#define RH_OFFSET 900

// --------------------------------------------------------------------------------------
//  DIODES Vf and TRANSISTORS vBE
// --------------------------------------------------------------------------------------
// examples to increase the measured values: 1.005, 1.001, 1.025
// examples to decrease the measured values: 0.995, 0.983, 0.97
#define DIODE_VF_CORRECTION 0.973
#define TRANSISTOR_NPN_CORRECTION 0.97
#define TRANSISTOR_PNP_CORRECTION 0.97

// --------------------------------------------------------------------------------------
//  BEEP
// --------------------------------------------------------------------------------------
// Resistance value for the hardware beep optionally connected to pin D10
#define RES_FOR_BEEP 2  // normally 2 or 3 ohm

// --------------------------------------------------------------------------------------
// AUTO REF, AUTO SCALE and BAUD RATE
// --------------------------------------------------------------------------------------
// The WITH_AUTO_REF option enables reading of internal REF-voltage 
//  to get factors for the Capacity measuring.
#define WITH_AUTO_REF

// The AUTOSCALE_ADC option enables the autoscale ADC (ADC use VCC and Bandgap Ref).
#define AUTOSCALE_ADC

// Usual baud rate is 500000 (sperimented as the best to reduce transmission errors)
#define BAUD_RATE 500000

// ======================================================================================
//  END OF CONFIG OPTIONS
// ======================================================================================

// The Nano CPU frequency is 16 MHz - The #define F_CPU 16000000UL is already defined

// ADC port, that is directly connected to the probes.
// The lower pins of this Port must be used for measurements.
// Please don't change the definitions of TP1, TP2 and TP3!
#define ADC_PORT PORTC
#define ADC_DDR DDRC
#define ADC_PIN PINC
#define TP1 0
#define TP2 1
#define TP3 2

// The TPREF pin can be connected with a 2.5V precision voltage reference
// Port pin for 2.5V precision reference used for VCC check (optional)
#define TPREF 4

// -------------------------------------- V_VCC defines the VCC Voltage of the ATmega in mV units
#define V_VCC 5000
// integer factors are used to change the ADC-value to mV resolution in ReadADC !

// With the option NO_CAP_HOLD_TIME you specify, that capacitor loaded with 680 Ohm resistor will not
// be tested to hold the voltage same time as load time.
// Otherwise (without this option) the voltage drop during load time is compensated to avoid displaying
// too much capacity for capacitors with internal parallel resistance.
// #define NO_CAP_HOLD_TIME

// V_SCALE can be set to 4 for better resolution of ReadADC function for resistor measurement
#define V_SCALE 4   // Please set this 1 to 4

// The N_MEASURES option specifies, how often an ADC value is read and accumulated.
// Possible values of N_MEASURES are 5 to 200 (normally 25)
#define N_MEASURES 25

// R_N_MEASURES can be set to a higher number of measurements (up to 200) for resistor measurement
#define R_N_MEASURES 190  // please set this more or equal then N_MEASURES

// ---------------------------------- selections for processor mega328
#define MCV_STATUS_REG MCUCR
#define ADC_COMP_CONTROL ADCSRB
#define TI1_INT_FLAGS TIFR1
#define DEFAULT_BAND_GAP 1070
#define DEFAULT_RH_FAKT  884 // mega328 1070 mV  <<< Multiplier for measurement with R_H (470KOhm)

// Typical resistance values for the mega328 pins switched to GND and to VCC
#define PIN_RM 200  // 200 means 20 ohm
#define PIN_RP 220  // 220 means 22 ohm

#define RR680PL (R_L_VAL + PIN_RP)
#define RR680MI (R_L_VAL + PIN_RM)
#define RRpinPL (PIN_RP)
#define RRpinMI (PIN_RM)

// CC0 defines the capacity of empty terminal pins 1 & 3 without cable
#define CC0 36

// Slew rate correction  val += COMP_SLEW1 / (val + COMP_SLEW2)
#define COMP_SLEW1 4000
#define COMP_SLEW2 180
#define C_NULL CC0 + CABLE_CAP + (COMP_SLEW1 / (CC0 + CABLE_CAP + COMP_SLEW2))
#define MUX_INT_REF 0x0e  // channel number of internal 1.1 V

// ----------------------------------- with Nano we use a 1.1V reference
#define ADC_internal_reference (ref_mv + REF_R_KORR)

// ----------------------------------------------------------------------------
// The ADC_CLOCK_PRESCALER defines the correct ADC-Clock 
// We set the ADC-Clock to 125000 (max for full accuracy specification)
// With CPV_Clock=16MHz and ADC_Clock=125kHz the clock divider is 128
// And 128 is specified with the ADC_CLOCK_PRESCALER is 4 + 2 + 1 = 7
// ----------------------------------------------------------------------------
#define ADC_CLOCK_PRESCALER 7
// ----------------------------------------------------------------------------

#define TXD_MSK 0xF8
#define TXD_VAL TXD_MSK

// The voltage at a capacitor grows with  Uc = VCC * (1 - e**(-t/T))
// The voltage 1.3V is reached at  t = -ln(3.7/5)*T  = 0.3011*T . 
// Time constant is  T = R * C ; also
// C = T / R
// for the resistor 470 kOhm  is C = t / (0.3011 * 470000)
// H_Fakt = 707/100 for a result in pF units.

// Big Capacities (>50uF) are measured with up to 500 load-pulses with the 680 Ohm resistor.
// Each  of this load-pulse has a length of 10ms. After every load-pulse the voltage of the
// capacitor is measured. If the voltage is more than 300mV, the capacity is computed by
// interpolating the corresponding values of the table RLtab and multiply that with the number
// of load pulses (*10).

// Widerstand 680 Ohm               300   325   350   375   400   425   450   475   500   525   550   575   600   625   650   675   700   725   750   775   800   825   850   875   900   925   950   975  1000  1025  1050  1075  1100  1125  1150  1175  1200  1225  1250  1275  1300  1325  1350  1375  1400  mV
const uint16_t RLtab[] PROGMEM = {22447,20665,19138,17815,16657,15635,14727,13914,13182,12520,11918,11369,10865,10401, 9973, 9577, 9209, 8866, 8546, 8247, 7966, 7702, 7454, 7220, 6999, 6789, 6591, 6403, 6224, 6054, 5892, 5738, 5590, 5449, 5314, 5185, 5061, 4942, 4828, 4718, 4613, 4511, 4413, 4319, 4228};

// table for function -1000*log(1 - (permil/1000)) , distance is 20 permil
//                                {0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,  32,  33,  34,  35,  36,  37,  38,  39,  40,  41,  42,  43,  44,  45,  46,  47,  48,  49,  50};
const uint16_t LogTab[] PROGMEM = {0, 20, 41, 62, 83,105,128,151,174,198,223,248,274,301,329,357,386,416,446,478,511,545,580,616,654,693,734,777,821,868,916,968,1022,1079,1139,1204,1273,1347,1427,1514,1609,1715,1833,1966,2120,2303,2526,2813,3219,3912,7065 }; /* last value is replaced by  (-1000*ln(0.001)-3912)/19*20 + 3912 */

#ifdef WITH_AUTO_REF
  // resistor  470000 Ohm             1000 1050 1100 1150 1200 1250 1300 1350 1400  mV
  //const uint16_t RHtab[] PROGMEM = { 954, 903, 856, 814, 775, 740, 707, 676, 648};

  const uint16_t RHtab[] PROGMEM = 
  { 
    uint16_t(44814201.0 / R_H_VAL + 0.5),   // 1000 mV
    uint16_t(42422794.0 / R_H_VAL + 0.5),   // 1050 mV
    uint16_t(40247707.0 / R_H_VAL + 0.5),   // 1100 mV
    uint16_t(38260704.0 / R_H_VAL + 0.5),   // 1150 mV
    uint16_t(36438256.0 / R_H_VAL + 0.5),   // 1200 mV
    uint16_t(34760595.0 / R_H_VAL + 0.5),   // 1250 mV
    uint16_t(33210996.0 / R_H_VAL + 0.5),   // 1300 mV
    uint16_t(31775210.0 / R_H_VAL + 0.5),   // 1350 mV
    uint16_t(30441023.0 / R_H_VAL + 0.5)    // 1400 mV
  };

  // these values are 1e7/R_H_VAL/-ln(1-ref/5000) (note that R_H_VAL is in multiples of 10 ohm)
  // the +0.5 is for rounding

  // with integer factors the ADC-value will be changed to mV resolution in ReadADC !
  // all if statements are corrected to the mV resolution.
#endif

// -------------------------------------- Strings 
#define AnKat " ->|- "
#define KatAn " -|<- "
#define Resistor_str " -==- " // Examples: " -==- " or " -[=]- " used for Resistors and Inductors 

const unsigned char PinRLtab[] PROGMEM = {(1<<(TP1*2)), (1<<(TP2*2)), (1<<(TP3*2))}; // Table of commands to switch the  R-L resistors Pin 0,1,2
const unsigned char PinADCtab[] PROGMEM = {(1<<TP1), (1<<TP2), (1<<TP3)}; // Table of commands to switch the ADC-Pins 0,1,2

const uint8_t PrefixTab[] PROGMEM = {'p', 'n' ,'u' ,'m' , 0, 'k', 'M'};  // p,n,u,m,-,k,M

// -------------------------------------- Multiplier for capacity measurement with R_H (470KOhm)
unsigned int RHmultip = DEFAULT_RH_FAKT;

struct Diode_t 
{
  uint8_t Anode;
  uint8_t Cathode;
  unsigned int Voltage;
};

struct Diode_t diodes[6];
uint8_t NumOfDiodes;

struct 
{
  unsigned long hfe[2];  // current amplification factor 
  unsigned int vBE[2];   // B-E-voltage of the Transistor
  uint8_t b,c,e;         // pins of the Transistor
} trans;

unsigned int gthvoltage; // Gate-threshold voltage 

uint8_t PartReady;       // part detection is finished 
uint8_t PartMode;
uint8_t tmpval, tmpval2;
unsigned int ref_mv;     // Reference-voltage  in mV units

struct resis_t
{
  unsigned long rx;      // value of resistor RX  
  unsigned long lx;      // inductance 10uH or 100uH
  int8_t lpre;           // prefix for inductance
  uint8_t ra,rb;         // Pins of RX
  uint8_t rt;            // Tristate-Pin (inactive)
} resis[3];

uint8_t ResistorsFound;  // Number of found resistors
uint8_t ii;              // multipurpose counter

struct cap_t 
{
  unsigned long cval;    // capacitor value 
  unsigned long cval_max;// capacitor with maximum value

  union t_combi
  {
    unsigned long dw;    // capacity value uncorrected
    uint16_t w[2];
  } cval_uncorrected;

  unsigned int esr;      // serial resistance of C in 0.01 Ohm
  unsigned int v_loss;   // voltage loss 0.1%
  
  uint8_t ca, cb;        // pins of capacitor
  int8_t cpre;           // Prefix for capacitor value  -12=p, -9=n, -6=u, -3=m
  int8_t cpre_max;       // Prefix of the biggest capacitor
} cap;

int16_t load_diff;       // difference voltage of loaded capacitor and internal reference

uint8_t WithReference;   // Marker for found precision voltage reference = 1
uint8_t PartFound;       // the found part 
uint8_t empty_count;     // counter for max count of empty measurements
uint8_t mess_count;      // counter for max count of nonempty measurements

struct ADCconfig_t 
{
  uint8_t Samples;       // number of ADC samples to take
  uint8_t RefFlag;       // save Reference type VCC of IntRef
  uint16_t V_Bandgap;    // Reference Voltage in mV
  uint16_t V_AVCC;       // Voltage of AVCC
} ADCconfig;

// -------------------------------- definitions of parts
#define PART_NONE 0
#define PART_DIODE 1
#define PART_TRANSISTOR 2
#define PART_FET 3
#define PART_TRIAC 4
#define PART_THYRISTOR 5
#define PART_RESISTOR 6
#define PART_CAPACITOR 7
#define PART_CELL 8

// -------------------------- special definition for different parts (FETs)
#define PART_MODE_N_E_MOS 2
#define PART_MODE_P_E_MOS 3
#define PART_MODE_N_D_MOS 4
#define PART_MODE_P_D_MOS 5
#define PART_MODE_N_JFET 6
#define PART_MODE_P_JFET 7

// -------------------------- special definition for different parts (Bipolar)
#define PART_MODE_NPN 1
#define PART_MODE_PNP 2

uint8_t tmp = 0;

byte TestKey;
byte TestKeyPin = 17;  // A3


// ################################################################################
// ################################################################################
// ################################################################################
//  FAST RESISTOR MEASUREMENT AND BEEP --- VARIABLES INITIALIZATION
// ################################################################################
// ################################################################################
// ################################################################################

#ifdef START_AS_FAST_TESTER
  byte TestType = 1;     // Start as Fast-Resistor and Beep-Tester
#else
  byte TestType = 2;     // Start as normal Transistor-Tester
#endif

// ------------------------------------ Adc and resistors pins for Fast-Resistor and Beep-Tester
byte pinAdcA = 0;  
byte pinAdcB = 1;
#ifdef RESISTORS_ON_D2D7
  byte pinResLow1 = 2;
  byte pinResLow2 = 4;
#else
  byte pinResLow1 = 8;
  byte pinResLow2 = 10;
#endif

// ------------------------------------ Voltage and resistance values for Fast-Resistor and Beep-Tester
unsigned int millivoltSupply = V_VCC;

int32_t resL = (R_L_VAL + (PIN_RP + PIN_RM) / 2) / 10;  


// ################################################################################
// ################################################################################
// ################################################################################
//   R-PORT and R-DDR modifications
// ################################################################################
// ################################################################################
// ################################################################################

// In the original "Transistor Tester by Karl-Heinz K¨ubbeler" the resistor drive pins was:
//  PORTB(0 to 5) corresponding to the Arduino Nano pins D8, D9, D10, D11, D12, D13 

// But in the Arduino boards the pin D13 drives a LED, so users must remove a resistor.

// So in this version we changed the resistor drive pins to:
//  PORTD(2 to 7) corresponding to the Arduino Nano pins D2, D3, D4, D5, D6, D7
//  And we also avoid modifying the serial RX and TX pins D0 and D1

#ifdef RESISTORS_ON_D2D7
  inline uint8_t PDatGet()            {return PORTD >> 2;} 
  inline void    PDatSet(uint8_t Dat) {PORTD = Dat << 2 | (PORTD & 0b11);} 
  inline uint8_t PDirGet()            {return DDRD >> 2;} 
  inline void    PDirSet(uint8_t Dir) {DDRD = Dir << 2 | (DDRD & 0b11);} 
#else 
  inline uint8_t PDatGet()            {return PORTB;} 
  inline void    PDatSet(uint8_t Dat) {PORTB = Dat;} 
  inline uint8_t PDirGet()            {return DDRB;} 
  inline void    PDirSet(uint8_t Dir) {DDRB = Dir;} 
#endif


// ################################################################################
// ################################################################################
// ################################################################################
//  SETUP
// ################################################################################
// ################################################################################
// ################################################################################

void setup()
{
  Serial.begin(BAUD_RATE);

  // ADC-Init
  ADCSRA = (1<<ADEN) | ADC_CLOCK_PRESCALER; // prescaler=8 or 64 (if 8Mhz clock)

  empty_count = 0;
  mess_count = 0;
}

// ################################################################################
// ################################################################################
// ################################################################################
//  FAST RESISTOR MEASUREMENT AND BEEP
// ################################################################################
// ################################################################################
// ################################################################################

// ================================================================================
//  TEST RESISTANCE
// ================================================================================
int32_t rawVoltagePos;
int32_t rawVoltageNeg;
int32_t microvolt;
int32_t milliohm;

void TestRES()
{
  TestResistance();
  BeepWithLowRes();
  PrintValuesToSerial();
}

void TestResistance()
{
  pinMode(pinResLow1, OUTPUT); 
  pinMode(pinResLow2, OUTPUT);  
  digitalWrite(pinResLow1, HIGH); 
  digitalWrite(pinResLow2, LOW);
  delayMicroseconds(100); 
  //
  TestVoltage();
  ComputeMilliohm(microvolt, resL);
  //
  pinMode(pinResLow1, INPUT); 
  pinMode(pinResLow2, INPUT); 
}

// ================================================================================
//  TEST VOLTAGE  
//  analogRead sampling time = about 100 uS
//  sampled 200 times (2 mS) and multiplied by 5 to get microvolts
// ================================================================================
void TestVoltage()
{
  rawVoltagePos = 0;
  rawVoltageNeg = 0;
  for (int i = 0; i < 200; i++)
  {
    rawVoltagePos += analogRead(pinAdcA); 
    rawVoltageNeg += analogRead(pinAdcB);
  }
  microvolt = 5 * int64_t(rawVoltagePos - rawVoltageNeg) * millivoltSupply / 1024;
  if (microvolt < 0) microvolt = 0;
}

// ================================================================================
//  COMPUTE RESISTANCE 
// ================================================================================
//     |     Voltage     | (SupplyVoltage - Voltage) |
//  0V ...... ResX .......... TestRes*2 .............. SupplyVoltage
// ----------------------------------------------------------------------------
//  Current = (SupplyVoltage - Voltage) / (TestRes * 2)
//  ResX = Voltage / Current
// ================================================================================
void ComputeMilliohm(int32_t microVolt, int32_t testRes)
{
  float currentMicroampere = (1000.0 * millivoltSupply - microVolt) / (testRes * 2.0);
  if (currentMicroampere < 10) currentMicroampere = 10;
  milliohm = 1000 * (microVolt / currentMicroampere);
}

// ================================================================================
//   PRINT VALUES TO SERIAL
// ================================================================================
void PrintValuesToSerial() 
{
  //Serial.print(String(milliohm) + " milliohm");  
  //Serial.println();

  Serial.println(String(milliohm) + " " + String(microvolt / 1000));
}

// ================================================================================
//   Read SERIAL DATA
// ================================================================================
void TestSerialData()
{
  if (Serial.available() > 0) 
  {
    char receivedData = Serial.read();
    //
    if (receivedData == '1')       // Test for low value resistors and Beep 
    {
      TestType = 1;
    } 
    else if (receivedData == '2')  // Test for LCR and active components
    {
      TestType = 2;
    } 
    else if (receivedData == '3')  // Start the test for LCR and active components
    {
      TestType = 3;
    }
  }
}

// ================================================================================
//  BEEP WITH LOW RES
// ================================================================================
void BeepWithLowRes()
{
  if (abs(milliohm) < RES_FOR_BEEP * 1000)
  {
    BeepPin10();
  }
  else
  {
    BeepStop();
  }
}

void BeepPin10()
{
  // ----------------------------- 1 KHz on Pin 9 or 10 (Timer 1 Fast PWM prescaler 64)
  unsigned long frequency = 1000;
  unsigned long period = F_CPU / (64UL * frequency) - 1;
  ICR1 = period;
  TCCR1A = (1 << WGM11) | (1 << COM1B1);
  TCCR1B = (1 << WGM13) | (1 << WGM12) | (1 << CS11) | (1 << CS10);
  analogWrite(10, 128); 
}

void BeepStop()
{
  pinMode(10, INPUT);
}

// void BeepMillisec(uint16_t millisec)
// {
//   BeepPin10();
//   delay(millisec);
//   BeepStop();
// }


// ################################################################################
// ################################################################################
// ################################################################################
//  LOOP
// ################################################################################
// ################################################################################
// ################################################################################

void loop()
{
  
start:

  // ################################################################################
  //  PREPARE FOR FAST RESISTOR MEASUREMENT AND BEEP 
  // ################################################################################
  MCUCR &= ~(1<<PUD);  // re-enable Pull-Up resistors
  pinMode(TestKeyPin, INPUT_PULLUP);
  // ################################################################################
   
  // ################################################################################
  //  IF TestType = 1 (or 3) then FAST RESISTOR MEASUREMENT AND BEEP
  // ################################################################################
  while (TestType != 2)
  {
    TestRES();
    TestSerialData();
  }
  BeepStop();

  // ################################################################################
  //  ELSE - WAIT START TEST COMMAND for SLOW TRANSISTOR TESTER 
  // ################################################################################
  TestKey = 1;
  while(TestKey == 1) 
  {
    delay(10);
    TestKey = digitalRead(TestKeyPin);
    TestSerialData();  
    if (TestType == 1) goto start; 
    if (TestType == 3) break; 
  }
  while(!TestKey) 
  {
    delay(10);
    TestKey = digitalRead(TestKeyPin);
    TestSerialData();  
    if (TestType == 1) goto start; 
    if (TestType == 3) break; 
  }
  if (TestType == 3) TestType = 2;

  // ################################################################################
  //  PREPARE FOR TRANSISTOR TESTER MESUREMENTS
  // ################################################################################
  MCUCR = (1<<PUD); // disable Pull-Up Resistors to remove any possible interference
  // ################################################################################
  
  Serial.println();
  delay(100); 

  PartFound = PART_NONE;  // no part found
  NumOfDiodes = 0;        // Number of diodes = 0
  PartReady = 0;
  PartMode = 0;
  WithReference = 0;      // no precision reference voltage
  ADC_DDR = TXD_MSK;      // activate Software-UART 
  ResistorsFound = 0;     // no resistors found
  cap.ca = 0;
  cap.cb = 0;

  ADCconfig.RefFlag = 0;
  Calibrate_VR();         // get Ref Voltages and Pin resistance
  Serial.println();       // 1 row
  
  ADCconfig.V_Bandgap = ADC_internal_reference;  // set internal reference voltage for ADC

  #ifdef WITH_AUTO_REF
    RefVoltage();         // compute RHmultip = f(reference voltage)
  #endif

  Serial.println();
  Serial.print("Testing..."); 

  // ================================================================================
  //   DISCHARGE CAPACITORS
  // ================================================================================
  delay(5);
  DischargePins();              // discharge capacitors

  if(PartFound == PART_CELL) 
  {
    Serial.println();
    Serial.print("Cell!");
    goto end2;
  }
     
  // ================================================================================
  //  Check all 6 combinations for the 3 pins 
  // ================================================================================
  //        High Low  Tri
  CheckPins(TP1, TP2, TP3);
  CheckPins(TP2, TP1, TP3);
  CheckPins(TP1, TP3, TP2);
  CheckPins(TP3, TP1, TP2);
  CheckPins(TP2, TP3, TP1);
  CheckPins(TP3, TP2, TP1);
  
  // separate check if it is a capacitor
  if(((PartFound == PART_NONE) || (PartFound == PART_RESISTOR) || (PartFound == PART_DIODE)) ) 
  {
    DischargePins();            // discharge capacities

    // measurement of capacities in all 3 combinations
    cap.cval_max = 0;           // set max to zero
    cap.cpre_max = -12;         // set max to pF unit

    ReadCapacity(TP3, TP1);
    ReadCapacity(TP3, TP2);
    ReadCapacity(TP2, TP1);

    ReadInductance();           // measure inductance
  }

  // ================================================================================
  //   All checks are done, output result to display
  // ================================================================================
  Serial.println();

  // ================================================================================
  //   PART FOUND - DIODE
  // ================================================================================
  if(PartFound == PART_DIODE) 
  {
    if(NumOfDiodes == 1)  // single Diode
    {   
      //Serial.print("Diode: "); 
      if (diodes[0].Anode < diodes[0].Cathode) 
      {
        ShowPinNumber(diodes[0].Anode);
        Serial.print(AnKat);             // "->|-"
        ShowPinNumber(diodes[0].Cathode);
      } 
      else 
      {
        ShowPinNumber(diodes[0].Cathode);
        Serial.print(KatAn);             // "-|<-"
        ShowPinNumber(diodes[0].Anode);
      }

      GetIr(diodes[0].Cathode,diodes[0].Anode);

      VfOutput(0x70);

      // load current of capacity is (5V-1.1V)/(470000 Ohm) = 8298nA
      Serial.print("C=");
      ReadCapacity(diodes[0].Cathode, diodes[0].Anode);  // Capacity opposite flow direction
      DisplayValue(cap.cval, cap.cpre, "F", 3);
      goto end;

    } 
    else if(NumOfDiodes == 2)                          // Double diode
    { 
      Serial.print("2");
      Serial.print(KatAn);             // "-|<-"

      if(diodes[0].Anode == diodes[1].Anode)           // Anode
      {
        ShowPinNumber(diodes[0].Cathode);
        Serial.print(KatAn);             // "-|<-"
        ShowPinNumber(diodes[0].Anode);
        Serial.print(AnKat);             // "->|-"
        ShowPinNumber(diodes[1].Cathode);
        VfOutput(0x01);
        goto end;
      } 
      else if(diodes[0].Cathode == diodes[1].Cathode)  // Cathode
      {
        ShowPinNumber(diodes[0].Anode);
        Serial.print(AnKat);             // "->|-"
        ShowPinNumber(diodes[0].Cathode);
        Serial.print(KatAn);             // "-|<-"
        ShowPinNumber(diodes[1].Anode);
        VfOutput(0x01);
        goto end;
      } 
      else if ((diodes[0].Cathode == diodes[1].Anode) && (diodes[1].Cathode == diodes[0].Anode))  // Antiparallel
      {
        ShowPinNumber(diodes[0].Anode);
        Serial.print(AnKat);             // "->|-"
        ShowPinNumber(diodes[0].Cathode);
        Serial.print(AnKat);             // "->|-"
        ShowPinNumber(diodes[1].Cathode);
        VfOutput(0x01);
        goto end;
      }

    } 
    else if(NumOfDiodes == 3) // Serial of 2 Diodes; was detected as 3 Diodes 
    {
      trans.b = 3;
      trans.c = 3;

      // Check for any constellation of 2 serial diodes:
      // Only once the pin No of anyone Cathode is identical of another anode.
      // two diodes in series is additionally detected as third big diode.

      if(diodes[0].Cathode == diodes[1].Anode)   { trans.b = 0; trans.c = 1; }
      if(diodes[0].Anode   == diodes[1].Cathode) { trans.b = 1; trans.c = 0; }
      if(diodes[0].Cathode == diodes[2].Anode)   { trans.b = 0; trans.c = 2; }
      if(diodes[0].Anode   == diodes[2].Cathode) { trans.b = 2; trans.c = 0; }
      if(diodes[1].Cathode == diodes[2].Anode)   { trans.b = 1; trans.c = 2; }
      if(diodes[1].Anode   == diodes[2].Cathode) { trans.b = 2; trans.c = 1; }

      if((trans.b < 3) && (trans.c < 3)) 
      {
        Serial.print("3");
        Serial.print(KatAn);          // "-|<-"
        ShowPinNumber(diodes[trans.b].Anode);
        Serial.print(AnKat);          // "->|-"
        ShowPinNumber(diodes[trans.b].Cathode);
        Serial.print(AnKat);          // "->|-"
        ShowPinNumber(diodes[trans.c].Cathode);
        VfOutput( (trans.b<<4)|trans.c);
        goto end;
      }
    }

  } // end (PartFound == PART_DIODE)

  // ================================================================================
  //   PART FOUND - TRANSISTOR
  // ================================================================================
  else if (PartFound == PART_TRANSISTOR) 
  {
    if(PartReady != 0)
    {
      if(trans.hfe[0] > trans.hfe[1]) 
      {
        // if the amplification factor was higher at first testr: swap C and E !
        tmp = trans.c;
        trans.c = trans.e;
        trans.e = tmp;
      } 
      else 
      {
        trans.hfe[0] = trans.hfe[1];
        trans.vBE[0] = trans.vBE[1];
      }
    }

    if(PartMode == PART_MODE_NPN) 
    {
      Serial.print("NPN ");
    } 
    else 
    {
      Serial.print("PNP "); 
    }

    if( NumOfDiodes > 2) // Transistor with protection diode
    { 
      if (((PartMode == PART_MODE_NPN) && (trans.c > trans.e)) || 
         ((PartMode != PART_MODE_NPN) && (trans.c < trans.e)))
      {
        Serial.print(AnKat);            // "->|-"
      } 
      else 
      {
        Serial.print(KatAn);            // "-|<-"
      }
    }

    PinLayout('E','B','C');             // EBC= or 123=...

    Serial.println();
    Serial.print("B=");
    DisplayValue(trans.hfe[0], 0, "", 3);
    Serial.print(" Vf=");
    DisplayValue(trans.vBE[0], -3, "V", 3);
    goto end;

  } // end (PartFound == PART_TRANSISTOR) 

  // ================================================================================
  //   PART FOUND - MOSFET // JFET or MOSFET
  // ================================================================================
  else if (PartFound == PART_FET) 
  { 
    if(PartMode&1) 
    {
      Serial.print("P");     // P-channel
    } else {
      Serial.print("N");     // N-channel
    }
    Serial.print("-");

    tmp = PartMode/2;
    if (tmp == (PART_MODE_N_D_MOS/2)) 
    {
      Serial.print("D");     // N-D
    }
    if (tmp == (PART_MODE_N_E_MOS/2)) 
    {
      Serial.print("E");     // N-E
    }

    if (tmp == (PART_MODE_N_JFET/2)) 
    {
      Serial.print("JFET");
    } 
    else 
    {
      Serial.print("-MOS");
    }

    PinLayout('S','G','D');             // SGD= or 123=...

    if((NumOfDiodes > 0) && (PartMode < PART_MODE_N_D_MOS)) {
      // MOSFET with protection diode; only with enhancement-FETs
      if (((PartMode&1) && (trans.c < trans.e)) || ((!(PartMode&1)) && (trans.c > trans.e)))
      {
        Serial.print(">|"); // show Diode symbol
      } else {
        Serial.print("|<"); // show Diode symbol 
      }
    }

    Serial.println();

    if(PartMode < PART_MODE_N_D_MOS)          // enhancement-MOSFET
    { 
      // Gate capacity
      Serial.print("C=");         
      ReadCapacity(trans.b, trans.e);         // measure capacity
      DisplayValue(cap.cval, cap.cpre, "F", 3);
      Serial.print(" Vt=");
    } 
    else 
    {
      Serial.print("I");
      Serial.print("=");
      DisplayValue(trans.vBE[1], -5, "A", 2);
      Serial.print("@Vgs="); 
    }

    // Gate-threshold voltage
    DisplayValue(gthvoltage, -3, "V", 2);
    goto end;

  }  // end (PartFound == PART_FET)
  
  // ================================================================================
  //   PART FOUND - THYRISTOR
  // ================================================================================
  else if (PartFound == PART_THYRISTOR) 
  {
    Serial.print("Thyristor");
    goto gakOutput;
  } 
  
  // ================================================================================
  //   PART FOUND - TRIAC
  // ================================================================================
  else if (PartFound == PART_TRIAC) 
  {
    Serial.print("Triac");
    goto gakOutput;
  } 
  
  // ================================================================================
  //   PART FOUND - RESISTOR
  // ================================================================================
  else if(PartFound == PART_RESISTOR) 
  {
    if (ResistorsFound == 1)               // single resistor
    {   
      ShowPinNumber(resis[0].rb);           // Pin-number 1
      Serial.print(Resistor_str);
      ShowPinNumber(resis[0].ra);           // Pin-number 2
    } 
    else 
    {                                      // find R-Max
      ii = 0;
      if (resis[1].rx > resis[0].rx) ii = 1;

      if (ResistorsFound == 2) 
      {
        ii = 2;
      } 
      else 
      {
        if (resis[2].rx > resis[ii].rx) ii = 2;
      }

      char x = '1';
      char y = '3';
      char z = '2';
   
      if (ii == 1) 
      {
        //x = '1';
        y = '2';
        z = '3';
      }

      if (ii == 2) 
      {
        x = '2';
        y = '1';
        z = '3';
      }

      Serial.write(x);
      Serial.print(Resistor_str);  
      Serial.write(y);
      Serial.print(Resistor_str); 
      Serial.write(z);
    }

    Serial.println();

    if (ResistorsFound == 1) 
    {
      RvalOut(0);
      if (resis[0].lx != 0)                               // resistor have also Inductance
      {
        Serial.print("L=");  
        DisplayValue(resis[0].lx, resis[0].lpre, "H", 3); // output inductance
      }
    } 
    else 
    {
      // output resistor values in right order
      if (ii == 0) 
      {
        RvalOut(1);
        RvalOut(2);
      }
      if (ii == 1) 
      {
        RvalOut(0);
        RvalOut(2);
      }
      if (ii == 2) 
      {
        RvalOut(0);
        RvalOut(1);
      }
    }
    goto end;
    
  }  // end (PartFound == PART_RESISTOR)
  
  // ================================================================================
  //   PART FOUND - CAPACITOR
  // ================================================================================
  else if(PartFound == PART_CAPACITOR) 
  {
    ShowPinNumber(cap.ca);               // Pin number 1
    Serial.print("-||-");               // capacitor sign
    ShowPinNumber(cap.cb);               // Pin number 2

    GetVloss();                         // get Voltage loss of capacitor
    if (cap.v_loss != 0) 
    {     
      Serial.print(" Vloss=");
      DisplayValue(cap.v_loss, -1, "%", 2);
    }

    Serial.println();
    DisplayValue(cap.cval_max, cap.cpre_max, "F", 4);

    // TODO
    // ------------------------ transform capacity to cap_val_nF
    unsigned long cap_val_nF;
    int8_t cpre = cap.cpre_max;
    cap_val_nF = cap.cval_max;
    while (cpre < -9)         // set cval to nF unit
    { 
      cap_val_nF /= 10;       // reduce value by factor ten
      cpre++;                 // take next decimal prefix
    }
    
    // Test the ESR for capacitors more than 900 nF
    if (cap_val_nF > 900)
    {
      cap.esr = GetESR(cap.cb, cap.ca); // get ESR of capacitor
 
      if (cap.esr > CAP_ESR_ZERO) cap.esr -= CAP_ESR_ZERO; else cap.esr = 0;

      Serial.print(" ESR=");
      DisplayValue(cap.esr, -2, "ohm", 2);
    }

    goto end;

  } // end (PartFound == PART_CAPACITOR)

  // ================================================================================
  //   UNKNOWN OR DAMAGED PART
  // ================================================================================
  if(NumOfDiodes == 0)              // no diodes are found
  {   
    Serial.print("No, unknown, or"); 
    Serial.println(); 
    Serial.print("damaged part");  
  } 
  else 
  {
    Serial.print("part unknown");
    Serial.println(); 
    Serial.print("or damaged ");
    Serial.write(NumOfDiodes + '0');
    Serial.print(AnKat);            // "->|-"
  }

  empty_count++;
  mess_count = 0;
  goto end2;

gakOutput:
  Serial.println(); 
  PinLayout('C','G','A');           // CGA= or 123=...

end:
  empty_count = 0;                  // reset counter, if part is found
  mess_count++;                     // count measurements

end2:
  //ADC_DDR = (1<<TPREF) | TXD_MSK; // switch pin with reference to GND, release relay
  ADC_DDR = TXD_MSK;                // switch pin with reference to GND, release relay

  Serial.println();
  
  goto start;  // repeat measurement

} // END LOOP



// ################################################################################
// ################################################################################
// ################################################################################
//  FUNCTIONS
// ################################################################################
// ################################################################################
// ################################################################################

// ====================================================================================
//  VfOutput - mVOutput
// ==================================================================================== 
// output of flux voltage for 1-2 diodes in row 2
// bcdnum = Numbers of both Diodes:
// higher 4 Bit  number of first Diode
// lower 4 Bit  number of second Diode (Structure diodes[nn])
// if number >= 3  no output is done
// ==================================================================================== 
void VfOutput(uint8_t bcdnum) 
{
  Serial.println();  
  Serial.print("Vf=");
  mVOutput(bcdnum >> 4);
  mVOutput(bcdnum & 0x0f);
}

void mVOutput(uint8_t nn) 
{
  if (nn < 3) 
  {
    // Output in mV units
    DisplayValue(diodes[nn].Voltage, -3, "V", 3);
    Serial.print(" ");
  }
}

// ====================================================================================
//  RVAL OUT - output of resistor value
// ==================================================================================== 
void RvalOut(uint8_t ii) 
{ 
  // ---------------------------------------------- If res < 100 ohm and not inductor
  if ((resis[ii].rx < 100) && (resis[0].lx == 0)) 
  {
    uint16_t rr = GetESR(resis[ii].ra, resis[ii].rb);       // Recalc with GetESR()
    if (rr > RES_ESR_ZERO) rr -= RES_ESR_ZERO; else rr = 0; // Subtract RES_ESR_ZERO
    DisplayValue(rr, -2, "ohm", 3);                         // And print
  } 
  else 
  {
    DisplayValue(resis[ii].rx, -1, "ohm", 4);
  }
  Serial.print(" ");
}

// ====================================================================================
//  CHARGE PIN 10 ms
// ==================================================================================== 
// Load the specified pin to the specified direction with 680 Ohm for 10ms.
// Will be used by discharge of MOSFET Gates or to load big capacities.
// Parameters:
// PinToCharge: specifies the pin as mask for R-Port
// ChargeDirection: 0 = switch to GND (N-Kanal-FET), 1= switch to VCC(P-Kanal-FET)
// ==================================================================================== 
void ChargePin10ms(uint8_t PinToCharge, uint8_t ChargeDirection) 
{
  if(ChargeDirection&1) 
  {
    PDatSet(PDatGet() | PinToCharge);   // switch Pin to 1 (VCC) 
  } else {
    PDatSet(PDatGet() & ~PinToCharge);  // or to 0 (GND)
  }
  PDirSet(PDirGet() | PinToCharge);     // switch Pin to output, across R to GND or VCC
  delay(10);                            // wait 10ms
  // switch back Input, no current
  PDirSet(PDirGet() & ~PinToCharge);    // switch back to input
  PDatSet(PDatGet() & ~PinToCharge);    // no Pull up
}

// ====================================================================================
//  DISCHARGE PINS - DISCHARGE CAPACITORS
// ==================================================================================== 
void DischargePins() 
{
  uint8_t adc_gnd;       // Mask of ADC-outputs, which can be directly connected to GND
  unsigned int adcmv[3]; // voltages of 3 Pins in mV
  unsigned int clr_cnt;  // Clear Counter
  uint8_t lop_cnt;       // loop counter

  // max. time of discharge  (10000/20) == 10s
  #define MAX_DISCHARGE_TIME  (10000/20)

  for(lop_cnt=0;lop_cnt<10;lop_cnt++) 
  {
    adc_gnd = TXD_MSK;    // put all ADC to Input
    ADC_DDR = adc_gnd;
    ADC_PORT = TXD_VAL;   // ADC-outputs auf 0
    PDatSet(0);           // R-outputs auf 0
    PDirSet((2<<(TP3*2)) | (2<<(TP2*2)) | (2<<(TP1*2)));  // R_H for all Pins to GND

    adcmv[0] = W5msReadADC(TP1); // which voltage has Pin 1?
    adcmv[1] = ReadADC(TP2);     // which voltage has Pin 2?
    adcmv[2] = ReadADC(TP3);     // which voltage has Pin 3?

    if ((PartFound == PART_CELL) || (adcmv[0] < CAP_EMPTY_LEVEL) 
                                  & (adcmv[1] < CAP_EMPTY_LEVEL) 
                                  & (adcmv[2] < CAP_EMPTY_LEVEL)) 
    {
      ADC_DDR = TXD_MSK;    // switch all ADC-Pins to input
      PDirSet(0);           // switch all R_L Ports (and R_H) to input
      return;               // all is discharged
    }

    // all Pins with voltage lower than 1V can be connected directly to GND (ADC-Port)
    if (adcmv[0] < 1000) 
    {
      adc_gnd |= (1<<TP1);  // Pin 1 directly to GND
    }
    if (adcmv[1] < 1000) 
    {
      adc_gnd |= (1<<TP2);  // Pin 2 directly to GND
    }
    if (adcmv[2] < 1000) 
    {
      adc_gnd |= (1<<TP3);  // Pin 3 directly to  GND
    }
    ADC_DDR = adc_gnd;      // switch all selected ADC-Ports at the same time

    // additionally switch the leaving Ports with R_L to GND.
    // since there is no disadvantage for the already directly switched pins, we can
    // simply switch all  R_L resistors to GND
    PDirSet((1<<(TP3*2)) | (1<<(TP2*2)) | (1<<(TP1*2)));  // Pins across R_L resistors to GND

    for(clr_cnt=0; clr_cnt < MAX_DISCHARGE_TIME; clr_cnt++) 
    {
      adcmv[0] = W20msReadADC(TP1); // which voltage has Pin 1?
      adcmv[1] = ReadADC(TP2);    // which voltage has Pin 2?
      adcmv[2] = ReadADC(TP3);    // which voltage has Pin 3?

      if (adcmv[0] < 1300) 
      {
         ADC_DDR |= (1<<TP1); // below 1.3V , switch directly with ADC-Port to GND
      }
      if (adcmv[1] < 1300) 
      {
         ADC_DDR |= (1<<TP2); // below 1.3V, switch directly with ADC-Port to GND
      }
      if (adcmv[2] < 1300) 
      {
         ADC_DDR |= (1<<TP3); // below 1.3V, switch directly with ADC-Port to GND
      }
      if ((adcmv[0] < (CAP_EMPTY_LEVEL+2)) && (adcmv[1] < (CAP_EMPTY_LEVEL+2)) && (adcmv[2] < (CAP_EMPTY_LEVEL+2))) 
      {
         break;
      }
    }

    if (clr_cnt == MAX_DISCHARGE_TIME)           // there is charge on capacitor, warn later!
    {
      PartFound = PART_CELL;                     // mark as Battery
    }

    for(adcmv[0]=0;adcmv[0]<clr_cnt;adcmv[0]++)  // for safety, discharge 5% of discharge time
    {
      delay(1);
    }
  }  // end for lop_cnt
}

// ====================================================================================
//  REF VOLTAGE
// ==================================================================================== 
// RefVoltage interpolates table RHtab corresponding to voltage ref_mv .
// RHtab contain the factors to get capacity from load time with 470k for
// different Band gab reference voltages.
// for remember:
// resistor    470000 Ohm      1000 1050 1100 1150 1200 1250 1300 1350 1400  mV
// uint16_t RHTAB[] PROGMEM = { 954, 903, 856, 814, 775, 740, 707, 676, 648};
// ==================================================================================== 
#ifdef WITH_AUTO_REF
  void RefVoltage(void) 
  {
    #define Ref_Tab_Abstand 50    // displacement of table is 50mV
    #define Ref_Tab_Beginn 1000   // begin of table is 1000mV

    unsigned int referenz;
    unsigned int y1, y2;
    uint8_t tabind;
    uint8_t tabres;
    
    referenz = ref_mv + REF_C_LOW_KORR;

    if (referenz >= Ref_Tab_Beginn) 
    {
      referenz -= Ref_Tab_Beginn;
    } 
    else 
    {
      referenz = 0; // limit to begin of table
    }

    tabind = referenz / Ref_Tab_Abstand;
    tabres = referenz % Ref_Tab_Abstand;
    tabres = Ref_Tab_Abstand-tabres;

    if (tabind > 7) 
    {
      tabind = 7;   // limit to end of table
    }

    // interpolate the table of factors
    y1 = pgm_read_word(&RHtab[tabind]);
    y2 = pgm_read_word(&RHtab[tabind+1]);
    // RHmultip is the interpolated factor to compute capacity from load time with 470k
    RHmultip = ((y1 - y2) * tabres + (Ref_Tab_Abstand/2)) / Ref_Tab_Abstand + y2;
  }
#endif

// =============================================================================================
//  CHECK PINS
// =============================================================================================
// Function for checking the characteristic of a component with the following pin assignment 
//  parameters:
//  HighPin: Pin, which will be switched to VCC at the beginning
//  LowPin: Pin, which will be switch to GND at the beginning
//  TristatePin: Pin, which will be undefined at the beginning
//  TristatePin will be switched to GND and VCC also .
// =============================================================================================
void CheckPins(uint8_t HighPin, uint8_t LowPin, uint8_t TristatePin) 
{
  struct 
  {
    unsigned int lp_otr;
    unsigned int hp1;
    unsigned int hp2;
    unsigned int hp3;
    unsigned int lp1;
    unsigned int lp2;
    unsigned int tp1;
    unsigned int tp2;
  } adc;

  uint8_t LoPinRL;    // mask to switch the LowPin with R_L
  uint8_t LoPinRH;    // mask to switch the LowPin with R_H
  uint8_t TriPinRL;   // mask to switch the TristatePin with R_L
  uint8_t TriPinRH;   // mask to switch the TristatePin with R_H
  uint8_t HiPinRL;    // mask to switch the HighPin with RL
  uint8_t HiPinRH;    // mask to switch the HighPin with R_H
  uint8_t HiADCp;     // mask to switch the ADC port High-Pin
  uint8_t LoADCp;     // mask to switch the ADC port Low-Pin
  uint8_t HiADCm;     // mask to switch the ADC DDR port High-Pin
  uint8_t LoADCm;     // mask to switch the ADC DDR port Low-Pin
  uint8_t PinMSK;
  uint8_t ii;         // temporary variable
  unsigned int tmp16; // temporary variable
  int udiff;

  struct resis_t *thisR;
  unsigned long lrx1;
  unsigned long lirx1;
  unsigned long lirx2;

  // switch HighPin directls to VCC 
  // switch R_L port for LowPin to GND 
  // TristatePin remains switched to input , no action required 

  LoPinRL = pgm_read_byte(&PinRLtab[LowPin]);       // instruction for LowPin R_L
  LoPinRH = LoPinRL + LoPinRL;                      // instruction for LowPin R_H
  TriPinRL = pgm_read_byte(&PinRLtab[TristatePin]); // instruction for TristatePin R_L
  TriPinRH = TriPinRL + TriPinRL;                   // instruction for TristatePin R_H
  HiPinRL = pgm_read_byte(&PinRLtab[HighPin]);      // instruction for HighPin R_L
  HiPinRH = HiPinRL + HiPinRL;                      // instruction for HighPin R_H

  HiADCp = pgm_read_byte(&PinADCtab[HighPin]);      // instruction for ADC High-Pin 
  LoADCp = pgm_read_byte(&PinADCtab[LowPin]);       // instruction for ADC Low-Pin
  HiADCm = HiADCp | TXD_MSK;
  HiADCp |= TXD_VAL;
  LoADCm = LoADCp | TXD_MSK;
  LoADCp |= TXD_VAL;

  // ------------------------------- setting of Pins 
  PDatSet(0);         // resistor-Port outputs to 0
  PDirSet(LoPinRL);   // Low-Pin to output and across R_L to GND
  ADC_DDR = HiADCm;   // High-Pin to output
  ADC_PORT = HiADCp;  // High-Pin fix to Vcc

  // -------------------------------- for some MOSFET the gate (TristatePin) must be discharged
  ChargePin10ms(TriPinRL,0);        // discharge for N-Kanal
  adc.lp_otr = W5msReadADC(LowPin); // read voltage of Low-Pin 

  if (adc.lp_otr >= 977)            // no current now? 
  { 
    ChargePin10ms(TriPinRL,1);      // else: discharge for P-channel (Gate to VCC)
    adc.lp_otr = ReadADC(LowPin);   // read voltage of Low-Pin again
  }

  //if(adc.lp_otr > 92)  // there is some current without TristatePin current 
  if(adc.lp_otr > 455)   // there is more than 650uA current without TristatePin current 
  {   
    // Test if N-JFET or if self-conducting N-MOSFET
    PDirSet(LoPinRL | TriPinRH);    // switch R_H for Tristate-Pin (probably Gate) to GND
    adc.lp1 = W20msReadADC(LowPin); // measure voltage at the assumed Source 
    adc.tp1 = ReadADC(TristatePin); // measure Gate voltage
    PDatSet(TriPinRH);              // switch R_H for Tristate-Pin (probably Gate) to VCC
    adc.lp2 = W20msReadADC(LowPin); // measure voltage at the assumed Source again

    // If it is a self-conducting MOSFET or JFET, then must be: adc.lp2 > adc.lp1 
    if (adc.lp2>(adc.lp1+488)) 
    {
      if (PartFound != PART_FET) 
      {
        // measure voltage at the  Gate, differ between MOSFET and JFET
        ADC_PORT = TXD_VAL;
        ADC_DDR = LoADCm;               // Low-Pin fix to GND
        PDirSet(TriPinRH | HiPinRL);    // High-Pin to output
        PDatSet(TriPinRH | HiPinRL);    // switch R_L for High-Pin to VCC
        adc.lp2 = W20msReadADC(TristatePin); // read voltage of assumed Gate 

        if (adc.lp2>3911)               // MOSFET
        { 
          PartFound = PART_FET;         // N-Kanal-MOSFET
          PartMode = PART_MODE_N_D_MOS; // Depletion-MOSFET
        } 
        else 
        {                               // JFET (pn-passage between Gate and Source is conducting )
          PartFound = PART_FET;         // N-Kanal-JFET
          PartMode = PART_MODE_N_JFET;
        }

        //if ((PartReady == 0) || (adc.lp1 > trans.vBE[0])) 
        // there is no way to find out the right Source / Drain
        trans.vBE[0] = adc.lp1;
        gthvoltage = adc.lp1 - adc.tp1; // voltage GS (Source - Gate)
        trans.vBE[1] = (unsigned int)(((unsigned long)adc.lp1 * 1000) / RR680MI);  // Id 0.01mA
        trans.b = TristatePin;    // save Pin numbers found for this FET
        trans.c = HighPin;
        trans.e = LowPin;
      }
    }

    ADC_PORT = TXD_VAL;   // direct outputs to GND

    // Test, if P-JFET or if self-conducting P-MOSFET
    ADC_DDR = LoADCm;                // switch Low-Pin (assumed Drain) direct to GND,
                                     // R_H for Tristate-Pin (assumed Gate) is already switched to VCC
    PDirSet(TriPinRH | HiPinRL);     // High-Pin to output
    PDatSet(TriPinRH | HiPinRL);     // High-Pin across R_L to Vcc
    adc.hp1 = W20msReadADC(HighPin); // measure voltage at assumed Source 
    adc.tp1 = ReadADC(TristatePin);  // measure Gate voltage
    PDatSet(HiPinRL);                // switch R_H for Tristate-Pin (assumed Gate) to GND
    adc.hp2 = W20msReadADC(HighPin); // read voltage at assumed Source again

    // if it is a self-conducting P_MOSFET or P-JFET , then must be:  adc.hp1 > adc.hp2 
    if (adc.hp1>(adc.hp2+488)) 
    {
      if (PartFound != PART_FET) 
      {
        // read voltage at the Gate , to differ between MOSFET and JFET
        ADC_PORT = HiADCp;                   // switch High-Pin directly to VCC
        ADC_DDR = HiADCm;                    // switch High-Pin to output
        adc.tp2 = W20msReadADC(TristatePin); //read voltage at the assumed Gate 

        if (adc.tp2<977)                // MOSFET
        { 
          PartFound = PART_FET;         // P-Kanal-MOSFET
          PartMode = PART_MODE_P_D_MOS; // Depletion-MOSFET
        } 
        else 
        {                               // JFET (pn-passage between Gate and Source is conducting)
          PartFound = PART_FET;         // P-Kanal-JFET
          PartMode = PART_MODE_P_JFET;
        }

        gthvoltage = adc.tp1 - adc.hp1;   // voltage GS (Gate - Source)

        trans.vBE[1] = (unsigned int)(((unsigned long)(ADCconfig.V_AVCC - adc.hp1) * 1000) / RR680PL); // Id 0.01mA

        trans.b = TristatePin;            // save Pin numbers found for this FET
        trans.c = LowPin;
        trans.e = HighPin;
      }
    }
  }  // end component has current without TristatePin signal

  // set Pins again for circuit with Emitter PNP
  PDirSet(LoPinRL);   // switch R_L port for Low-Pin to output (GND)
  PDatSet(0);         // switch all resistor ports to GND
  ADC_DDR = HiADCm;   // switch High-Pin to output
  ADC_PORT = HiADCp;  // switch High-Pin to VCC
  delay(5);
  
  if (adc.lp_otr < 977) // if the component has no connection between  HighPin and LowPin
  {
    // Test for PNP
    PDirSet(LoPinRL | TriPinRL);    // switch R_L port for Tristate-Pin to output (GND), for Test of PNP
    adc.lp1 = W5msReadADC(LowPin);  // measure voltage at LowPin

    if (adc.lp1 > 3422) 
    {
      // component has current => PNP-Transistor or equivalent
      // compute current amplification factor in both directions
      PDirSet(LoPinRL | TriPinRH);    // switch R_H port for Tristate-Pin (Base) to output (GND)

      adc.lp1 = W5msReadADC(LowPin);  // measure voltage at LowPin (assumed Collector)
      adc.tp2 = ReadADC(TristatePin); // measure voltage at TristatePin (Base) 

      // check, if Test is done before 
      if ((PartFound == PART_TRANSISTOR) || (PartFound == PART_FET)) 
      {
         PartReady = 1;
      }

      // hfE for common emitter transistor
      trans.vBE[PartReady] = ReadADC(HighPin) - adc.tp2;  // Base Emitter Voltage
      trans.vBE[PartReady] *= TRANSISTOR_PNP_CORRECTION;  // vBE correction for PNP transistors

      // compute current amplification factor for circuit with Emitter
      // hFE = B = Collector current / Base current

      if(adc.tp2 < 53) 
      {
        adc.tp2 = 53;
      }
      tmp16 = adc.lp1;

      if (tmp16 > adc.lp_otr) 
      {
        tmp16 -= adc.lp_otr;
      }

      trans.hfe[PartReady] = ((unsigned int)tmp16 * (unsigned long)(((unsigned long)R_H_VAL * 100) / 
                                (unsigned int)RR680MI)) / (unsigned int)adc.tp2; 

 
      if (PartFound != PART_THYRISTOR) 
      {
        if (adc.tp2 > 977) // PNP-Transistor is found (Base voltage moves to VCC)
        {
          PartFound = PART_TRANSISTOR;
          PartMode = PART_MODE_PNP;
        } 
        else 
        {
          // is flow voltage low enough in the closed  state?
          // (since D-Mode-FET would be by mistake detected as E-Mode )
          if ((adc.lp_otr < 97) && (adc.lp1 > 2000)) 
          {
            PartFound = PART_FET;   // P-Kanal-MOSFET is found (Basis/Gate moves not to VCC)
            PartMode = PART_MODE_P_E_MOS;
    
            // measure the Gate threshold voltage
            // Switching of Drain is monitored with digital input
            // Low level is specified up to 0.3 * VCC
            // High level is specified above 0.6 * VCC
            PinMSK = LoADCm & 7;
            ADMUX = TristatePin | (1<<REFS0);  // switch to TristatePin, Ref. VCC
            gthvoltage = 1;                    // round up ((1*4)/9)
    
            for(ii=0; ii<11; ii++) 
            {
              ChargePin10ms(TriPinRL,1);
              PDirSet(LoPinRL | TriPinRH);     // switch R_H for Tristate-Pin (Basis) to GND
    
              while (!(ADC_PIN&PinMSK));       // Wait, until the MOSFET switches and Drain moves to VCC
                                               // 1 is detected with more than 2.5V (up to 2.57V) with tests of mega168 and mega328
              PDirSet(LoPinRL);
              ADCSRA |= (1<<ADSC);             // Start Conversion
    
              while (ADCSRA&(1<<ADSC));        // wait

              gthvoltage += (1023 - ADCW);     // Add Tristatepin-Voltage
            }
    
            gthvoltage *= 4;    // is equal to 44*ADCW
            gthvoltage /= 9;    // gives resolution in mV
          }
        }

        trans.b = TristatePin;
        trans.c = LowPin;
        trans.e = HighPin;

      }  // end if PartFound != PART_THYRISTOR
      
    }  // end component has current => PNP

    // Tristate (can be Base) to VCC, Test if NPN
    ADC_DDR = LoADCm;               // Low-Pin to output 0V
    ADC_PORT = TXD_VAL;             // switch Low-Pin to GND
    PDirSet(TriPinRL | HiPinRL);    // RL port for High-Pin and Tristate-Pin to output
    PDatSet(TriPinRL | HiPinRL);    // RL port for High-Pin and Tristate-Pin to Vcc
    adc.hp1 = W5msReadADC(HighPin); // measure voltage at High-Pin  (Collector)
    
    if (adc.hp1 < 1600)  // component has current => NPN-Transistor or somthing else
    {    
      if (PartReady==1) 
      {
        goto widmes;
      }
    
      // Test for Thyristor:
      // Gate discharge
      ChargePin10ms(TriPinRL,0);      // Tristate-Pin (Gate) across R_L 10ms to GND
      adc.hp3 = W5msReadADC(HighPin); // read voltage at High-Pin (probably Anode) again
                                      // current should still flow, if not,
                                      // no Thyristor or holding current to low 
    
      PDatSet(0);                     // switch R_L for High-Pin (probably Anode) to GND (turn off)
      delay(5);
      PDatSet(HiPinRL);               // switch R_L for High-Pin (probably Anode) again to VCC
      adc.hp2 = W5msReadADC(HighPin); // measure voltage at the High-Pin (probably Anode) again
    
      if ((adc.hp3 < 1600) && (adc.hp2 > 4400)) 
      {
        // if the holding current was switched off the thyristor must be switched off too. 
        // if Thyristor was still swiched on, if gate was switched off => Thyristor
        PartFound = PART_THYRISTOR;
    
        // Test if Triac
        PDirSet(0);
        PDatSet(0);
        ADC_PORT = LoADCp;           // Low-Pin fix to VCC
        delay(5);
    
        PDirSet(HiPinRL);            // switch R_L port HighPin to output (GND)
        if(W5msReadADC(HighPin) > 244) 
        {
          goto savenresult;          // measure voltage at the  High-Pin (probably A2); if too high:
                                     // component has current => kein Triac
        }
    
        PDirSet(HiPinRL | TriPinRL); // switch R_L port for TristatePin (Gate) to output (GND) => Triac should be triggered 
        if(W5msReadADC(TristatePin) < 977)
        {
          goto savenresult;          // measure voltage at the Tristate-Pin (probably Gate) ;
                                     // if to low, abort 
        }
    
        if(ReadADC(HighPin) < 733) 
        {
          goto savenresult;     // component has no current => no Triac => abort
        }
    
        PDirSet(HiPinRL);       // TristatePin (Gate) to input 
        if(W5msReadADC(HighPin) < 733) 
        {
          goto savenresult;     // component has no current without base current => no Triac => abort
        }
    
        PDatSet(HiPinRL);       // switch R_L port for HighPin to VCC => switch off holding current 
        delay(5);
        PDatSet(0);             // switch R_L port for HighPin again to GND; Triac should now switched off
        if(W5msReadADC(HighPin) > 244) 
        {
          goto savenresult;     // measure voltage at the High-Pin (probably A2) ;
                                // if to high, component is not switched off => no Triac, abort
        }
    
        PartFound = PART_TRIAC;
        PartReady = 1;
        goto savenresult;
      }
    
      // Test if NPN Transistor or MOSFET
      //ADC_DDR = LoADCm;   // Low-Pin to output 0V
      PDirSet(HiPinRL | TriPinRH);  // R_H port of Tristate-Pin (Basis) to output
      PDatSet(HiPinRL | TriPinRH);  // R_H port of Tristate-Pin (Basis) to VCC
      delay(50);
      adc.hp2 = ADCconfig.V_AVCC - ReadADC(HighPin);     // measure the voltage at the collector resistor 
      adc.tp2 = ADCconfig.V_AVCC - ReadADC(TristatePin); // measure the voltage at the base resistor 
        
      if((PartFound == PART_TRANSISTOR) || (PartFound == PART_FET)) 
      {
        PartReady = 1;    // check, if test is already done once
      }
    
      // hfE for common emitter NPN transistor

      //  compute VBE for NPN transistors
      trans.vBE[PartReady] = ADCconfig.V_AVCC - adc.tp2 - ReadADC(LowPin);
      trans.vBE[PartReady] *= TRANSISTOR_NPN_CORRECTION; // vBE correction for NPN transistors
      
      // compute current amplification factor hFE = B = Collector current / Base current
      if (adc.tp2 < 53) 
      {    
        adc.tp2 = 53;
      }
  
      tmp16 = adc.hp2;
      if (tmp16 > adc.lp_otr) 
      {
        tmp16 -= adc.lp_otr;
      }
  
      trans.hfe[PartReady] = ((unsigned int)tmp16 * (unsigned long)(((unsigned long)R_H_VAL * 100) / 
                                (unsigned int)RR680PL)) / (unsigned int)adc.tp2; 
    
        
      if(adc.tp2 > 2557)
      {   // Basis-voltage R_H is low enough
        PartFound = PART_TRANSISTOR;  // NPN-Transistor is found (Base is near GND)
        PartMode = PART_MODE_NPN;
      } 
      else 
      {       // Basis has low current
        if((adc.lp_otr < 97) && (adc.hp2 > 3400)) 
        {
          // if flow voltage in switched off mode low enough?
          // (since D-Mode-FET will be detected in error as E-Mode )
          PartFound = PART_FET;   // N-Kanal-MOSFET is found (Basis/Gate will Not be pulled down)
          PartMode = PART_MODE_N_E_MOS;
        
          // Switching of Drain is monitored with digital input
          // Low level is specified up to 0.3 * VCC
          // High level is specified above 0.6 * VCC
          PinMSK = HiADCm & 7;
    
          // measure Threshold voltage of Gate
          ADMUX = TristatePin | (1<<REFS0); // measure TristatePin, Ref. VCC
          gthvoltage = 1;     // round up ((1*4)/9)
    
          for(ii=0; ii<11; ii++) 
          {
            ChargePin10ms(TriPinRL,0);   // discharge Gate 10ms with RL 
            PDirSet(HiPinRL | TriPinRH); // slowly charge Gate 
            PDatSet(HiPinRL | TriPinRH);
            while ((ADC_PIN&PinMSK));    // Wait, until the MOSFET switch and Drain moved to low 
                                         // 0 is detected with input voltage of 2.12V to 2.24V (tested with mega168 & mega328)
            PDirSet(HiPinRL);            // switch off current
            ADCSRA |= (1<<ADSC);         // start ADC conversion
            while (ADCSRA&(1<<ADSC));    // wait until ADC finished
            gthvoltage += ADCW;          // add result of ADC
          }
    
          gthvoltage *= 4;  // is equal to 44 * ADCW
          gthvoltage /= 9;  // scale to mV
        }
      }

savenresult:
      trans.b = TristatePin;  // save Pin-constellation
      trans.c = HighPin;
      trans.e = LowPin;
    }  // end component conduct => npn

    ADC_DDR = TXD_MSK;    // switch all ADC-Ports to input
    ADC_PORT = TXD_VAL;   // switch all ADC-Ports to 0 (no Pull up)

    // Finish
    // end component has no connection between HighPin and LowPin
    goto widmes;
  }

  // component has current
  // Test if Diode
  ADC_PORT = TXD_VAL;

  for (ii=0; ii<200; ii++) 
  {
    ADC_DDR = LoADCm | HiADCm;    // discharge by short of Low and High side
    delay(5);                     // Low and Highpin to GND for discharge
    ADC_DDR = LoADCm;             // switch only Low-Pin fix to GND
    adc.hp1 = ReadADC(HighPin);   // read voltage at High-Pin
    if (adc.hp1 < (150/8)) break;
  }

  // It is possible, that wrong Parts are detected without discharging, because
  // the gate of a MOSFET can be charged.
  // The additional measurement with the big resistor R_H is made, to differ antiparallel diodes
  // from resistors.
  // A diode has a voltage, that is nearly independent from the current.
  // The voltage of a resistor is proportional to the current.

  #if 0
    // first check with higher current (R_L=680)
    // A diode is found better with a parallel mounted capacitor,
    // but some capacitors can be detected as a diode.
    PDirSet(HiPinRL);               // switch R_L port for High-Pin to output (VCC)
    PDatSet(HiPinRL);
    ChargePin10ms(TriPinRL,1);      // discharge of P-Kanal-MOSFET gate
    adc.lp_otr = W5msReadADC(HighPin) - ReadADC(LowPin);
    PDirSet(HiPinRH);               // switch R_H port for High-Pin output (VCC)
    PDatSet(HiPinRH);
    adc.hp2 = W5msReadADC(HighPin); // M--|<--HP--R_H--VCC
    
    PDirSet(HiPinRL);               // switch R_L port for High-Pin to output (VCC)
    PDatSet(HiPinRL);
    ChargePin10ms(TriPinRL,0);      // discharge for N-Kanal-MOSFET gate
    adc.hp1 = W5msReadADC(HighPin) - W5msReadADC(LowPin);
    PDirSet(HiPinRH);               // switch R_H port for High-Pin to output (VCC)
    PDatSet(HiPinRH);
    adc.hp3 = W5msReadADC(HighPin); // M--|<--HP--R_H--VCC

    if(adc.lp_otr > adc.hp1) 
    {
      adc.hp1 = adc.lp_otr;         // the higher value wins
      adc.hp3 = adc.hp2;
    }
  #else
    // check first with low current (R_H=470k)
    // With this method the diode can be better differed from a capacitor,
    // but a parallel to a capacitor mounted diode can not be found.
    PDirSet(HiPinRH);     // switch R_H port for High-Pin output (VCC)
    PDatSet(HiPinRH);
    ChargePin10ms(TriPinRL,1);      // discharge of P-Kanal-MOSFET gate
    adc.hp2 = W5msReadADC(HighPin); // M--|<--HP--R_H--VCC
    ChargePin10ms(TriPinRL,0);      // discharge for N-Kanal-MOSFET gate
    adc.hp3 = W5msReadADC(HighPin); // M--|<--HP--R_H--VCC
    
    // check with higher current (R_L=680)
    PDirSet(HiPinRL);               // switch R_L port for High-Pin to output (VCC)
    PDatSet(HiPinRL);
    adc.hp1 = W5msReadADC(HighPin) - ReadADC(LowPin);

    ChargePin10ms(TriPinRL,1);      // discharge for N-Kanal-MOSFET gate
    adc.lp_otr = W5msReadADC(HighPin) - ReadADC(LowPin);
    
    PDirSet(HiPinRH);               // switch R_H port for High-Pin output (VCC)
    PDatSet(HiPinRH);

    if(adc.lp_otr > adc.hp1) 
    {
      adc.hp1 = adc.lp_otr;         // the higher value wins
      adc.hp3 = adc.hp2;
    }
    else 
    {
      ChargePin10ms(TriPinRL,0);    // discharge for N-Kanal-MOSFET gate
    }

    adc.hp2 = W5msReadADC(HighPin); // M--|<--HP--R_H--VCC
  #endif

  //if((adc.hp1 > 150) && (adc.hp1 < 4640) && (adc.hp1 > (adc.hp3+(adc.hp3/8))) && (adc.hp3*8 > adc.hp1)) 
  if((adc.hp1 > 150) && (adc.hp1 < 4640) && (adc.hp2 < adc.hp1) && (adc.hp1 > (adc.hp3+(adc.hp3/8))) && (adc.hp3*16 > adc.hp1)) 
  {
    // voltage is above 0,15V and below 4,64V => Ok
    if((PartFound == PART_NONE) || (PartFound == PART_RESISTOR)) 
    {
      PartFound = PART_DIODE; // mark for diode only, if no other component is found
                              // since there is a problem with Transistors with a protection diode
    }

    diodes[NumOfDiodes].Anode = HighPin;
    diodes[NumOfDiodes].Cathode = LowPin;

    // **********************************************************************************************************
    //   Vf correction for DIODES
    // **********************************************************************************************************
    diodes[NumOfDiodes].Voltage = adc.hp1;               // voltage in Millivolt   
    diodes[NumOfDiodes].Voltage *= DIODE_VF_CORRECTION;
    // **********************************************************************************************************

    NumOfDiodes++;
  }  // end voltage is above 0,15V and below 4,64V 

widmes:
  if (NumOfDiodes > 0) goto clean_ports;
  
  // resistor measurement

  // V_SCALE can be set to 4 for better resolution of ReadADC result
  #if V_SCALE != 1
    ADCconfig.V_AVCC *= V_SCALE;      // scale to higher resolution, mV scale is not required
    ADCconfig.V_Bandgap *= V_SCALE;
  #endif

  #if R_N_MEASURES != N_MEASURES
    ADCconfig.Samples = R_N_MEASURES; // switch to special number of repetitions
  #endif

  #define MAX_REPEAT (700 / (5 + R_N_MEASURES/8))

  ADC_PORT = TXD_VAL;
  ADC_DDR = LoADCm;   // switch Low-Pin to output (GND)
  PDirSet(HiPinRL);   // switch R_L port for High-Pin to output (VCC)
  PDatSet(HiPinRL); 

  adc.hp2 = 0;

  for (ii=1; ii<MAX_REPEAT; ii++)   // wait until voltage is stable
  {
    adc.tp1 = W5msReadADC(LowPin);  // low-voltage at Rx with load
    adc.hp1 = ReadADC(HighPin);     // voltage at resistor Rx with R_L
    udiff = adc.hp1 - adc.hp2;
    if (udiff < 0) udiff = -udiff;
    if (udiff < 3) break;
    adc.hp2 = adc.hp1;
  }

  if (ii == MAX_REPEAT) goto testend;

  if (adc.tp1 > adc.hp1) 
  {
    adc.tp1 = adc.hp1;
  }

  PDatSet(0);
  PDirSet(HiPinRH);               // switch R_H port for High-Pin to output (GND)
  adc.hp2 = W5msReadADC(HighPin); // read voltage, should be down

  if (adc.hp2 > (20 * V_SCALE)) 
  {
    // if resistor, voltage should be down
    goto testend;
  }

  PDatSet(HiPinRH);                 // switch R_H for High-Pin to VCC
  adc.hp2 = W5msReadADC(HighPin);   // voltage at resistor Rx with R_H

  ADC_DDR = HiADCm;                 // switch High-Pin to output
  ADC_PORT = HiADCp;                // switch High-Pin to VCC
  PDatSet(0);
  PDirSet(LoPinRL);                 // switch R_L for Low-Pin to GND

  adc.lp2 = 0;

  for (ii=1; ii<MAX_REPEAT; ii++)   // wait until voltage is stable
  {
    adc.tp2 = W5msReadADC(HighPin); // high voltage with load
    adc.lp1 = ReadADC(LowPin);      // voltage at the other end of Rx
    udiff = adc.lp1 - adc.lp2;
    if (udiff < 0) udiff = -udiff;
    if (udiff < 3) break;
    adc.lp2 = adc.lp1;
  }

  if (ii == MAX_REPEAT) goto testend;

  if (adc.tp2 < adc.lp1) 
  {
    adc.tp2 = adc.lp1;
  }

  PDirSet(LoPinRH);     // switch R_H for Low-Pin to GND
  adc.lp2 = W5msReadADC(LowPin);

  if((adc.hp1 < (4400 * V_SCALE)) && (adc.hp2 > (97 * V_SCALE))) 
  {
    //voltage break down isn't insufficient 
    goto testend; 
  }

  //if ((adc.hp2 + (adc.hp2 / 61)) < adc.hp1)
  if (adc.hp2 < (4972 * V_SCALE)) 
  { 
    // voltage breaks down with low test current and it is not nearly shorted  => resistor
    //if (adc.lp1 < 120)              // take measurement with R_H 
    if (adc.lp1 < (169 * V_SCALE))    // take measurement with R_H 
    {  
      ii = 'H';

      if (adc.lp2 < (38 * V_SCALE))   // measurement > 60 MOhm to big resistance 
      {
        goto testend;
      }

      // two measurements with R_H resistors (470k) are made:
      // lirx1 (measurement at HighPin)
      lirx1 = (unsigned long)((unsigned int)R_H_VAL) * (unsigned long)adc.hp2 / (ADCconfig.V_AVCC - adc.hp2);
      // lirx2 (measurement at LowPin)
      lirx2 = (unsigned long)((unsigned int)R_H_VAL) * (unsigned long)(ADCconfig.V_AVCC - adc.lp2) / adc.lp2;

      #define V_INT_LIMIT (990 * V_SCALE)   // 1V switch limit in ReadADC for atmega family

      #define FAKT_LOW 4    // resolution is about four times better

      #ifdef AUTOSCALE_ADC
        if (adc.hp2 < V_INT_LIMIT) 
        {
          lrx1 = (lirx1*FAKT_LOW + lirx2) / (FAKT_LOW+1); // weighted average of both R_H measurements
        } 
        else if (adc.lp2 < V_INT_LIMIT)
        {
          lrx1 = (lirx2*FAKT_LOW + lirx1) / (FAKT_LOW+1); // weighted average of both R_H measurements
        } else 
      #endif
        {
          lrx1 = (lirx1 + lirx2) / 2; // average of both R_H measurements
        }

      lrx1 *= 100;
      lrx1 += RH_OFFSET * 10;   // add constant for correction of systematic error

    }
    else 
    {
      ii = 'L';
      // two measurements with R_L resistors (680) are made:
      // lirx1 (measurement at HighPin)

      if (adc.tp1 > adc.hp1)    // negative diff is illegal
      {
        adc.hp1 = adc.tp1;
      }

      lirx1 =(unsigned long)RR680PL * (unsigned long)(adc.hp1 - adc.tp1) / (ADCconfig.V_AVCC - adc.hp1);

      if (adc.tp2 < adc.lp1)    // negative diff is illegal
      {
        adc.lp1 = adc.tp2;
      }

      // lirx2 (Measurement at LowPin)
      lirx2 =(unsigned long)RR680MI * (unsigned long)(adc.tp2 -adc.lp1) / adc.lp1;
      //lrx1 =(unsigned long)R_L_VAL * (unsigned long)adc.hp1 / (adc.hp3 - adc.hp1);

      #ifdef AUTOSCALE_ADC
        if (adc.hp1 < V_INT_LIMIT) 
        {
          lrx1 = (lirx1*FAKT_LOW + lirx2) / (FAKT_LOW+1); // weighted average of both R_L measurements
        } 
        else if (adc.lp1 < V_INT_LIMIT) 
        {
          lrx1 = (lirx2*FAKT_LOW + lirx1) / (FAKT_LOW+1); // weighted average of both R_L measurements
        } else
      #endif
        {
          lrx1 = (lirx1 + lirx2) / 2; // average of both R_L measurements
        }
    }
    // lrx1  is tempory result

    if((PartFound == PART_DIODE) || (PartFound == PART_NONE) || (PartFound == PART_RESISTOR)) 
    {
      for (ii=0; ii<ResistorsFound; ii++) 
      {
        // search measurements with inverse polarity 
        thisR = &resis[ii];
        if (thisR->rt != TristatePin) continue;

        // must be measurement with inverse polarity 
        // resolution is 0.1 Ohm, 1 Ohm = 10 !
        lirx1 = (labs((long)lrx1 - (long)thisR->rx) * 10) / (lrx1 + thisR->rx + 100);

        if (lirx1  > 0) 
        {
          goto testend;  // <10% mismatch
        }

        PartFound = PART_RESISTOR;
        goto testend;
      }  // end for

      // no same resistor with the same Tristate-Pin found, new one
      thisR = &resis[ResistorsFound]; // pointer to a free resistor structure
      thisR->rx = lrx1;        // save resistor value

      thisR->lx = 0;           // no inductance

      thisR->ra = LowPin;      // save Pin numbers
      thisR->rb = HighPin;
      thisR->rt = TristatePin; // Tristate is saved for easier search of inverse measurement
      ResistorsFound++;        // 1 more resistor found
    }
  }

testend:

  #if V_SCALE != 1
    ADCconfig.V_AVCC /= V_SCALE;      // scale back to mV resolution
    ADCconfig.V_Bandgap /= V_SCALE;
  #endif

  #if R_N_MEASURES != N_MEASURES
    ADCconfig.Samples = N_MEASURES;   // switch back to standard number of repetition
  #endif

clean_ports:

  ADC_DDR = TXD_MSK;  // all ADC-Pins Input
  ADC_PORT = TXD_VAL; // all ADC outputs to Ground, keine Pull up
  PDirSet(0);         // all resistor-outputs to Input
  PDatSet(0);         // all resistor-outputs to Ground, no Pull up

}  // end CheckPins()



// ====================================================================================
//  Get residual current in reverse direction of a diode
// ====================================================================================
void GetIr(uint8_t hipin, uint8_t lopin) 
{
  unsigned int v_res;     // reverse voltage at 470k
  unsigned int ir_nano;
  uint8_t LoPinR_L;
  uint8_t HiADC;

  HiADC = pgm_read_byte(&PinADCtab[hipin]);
  ADC_PORT = HiADC | TXD_VAL;    // switch ADC port to high level
  ADC_DDR = HiADC | TXD_MSK;     // switch High Pin direct to VCC
  LoPinR_L = pgm_read_byte(&PinRLtab[lopin]);   // R_L mask for LowPin R_L load
  PDatSet(0);                    // switch R-Port to GND
  PDirSet(LoPinR_L + LoPinR_L);  // switch R_H port for LowPin to output (GND)

  v_res = W5msReadADC(lopin);    // read voltage
  if (v_res == 0) return;        // no Output, if no current in reverse direction

  // Show the reverse current (nA)
  Serial.print("  Ir=");  
  // R_H_VAL has units of 10 Ohm, v_res has units of mV, ir_nano has units of nA
  ir_nano = (unsigned long)(v_res * 100000UL) / R_H_VAL;
  DisplayValue(ir_nano, -9, "A", 2); // output two digits of current with nA units

  ADC_DDR = TXD_MSK;  // switch off
  ADC_PORT = TXD_VAL; // switch off
  PDirSet(0);         // switch off current

  return;
}

// ====================================================================================
//  DEFINE StartADCwait
// ====================================================================================
#define StartADCwait() ADCSRA = (1<<ADSC) | (1<<ADEN) | (1<<ADIF) | ADC_CLOCK_PRESCALER; /* enable ADC and start */\
                                                           while (ADCSRA & (1 << ADSC))  /* wait until conversion is done */



// ====================================================================================
//  READ ADC
// ====================================================================================
unsigned int ReadADC (uint8_t Probe) 
{
  unsigned int U;        // return value (mV)
  uint8_t Samples;       // loop counter
  unsigned long Value;   // ADC value
  Probe |= (1 << REFS0); // use internal reference anyway

  #ifdef AUTOSCALE_ADC
sample:
  #endif

  ADMUX = Probe;  // set input channel and U reference

  #ifdef AUTOSCALE_ADC
    // switch to 1.1V Reference
    // if voltage reference changes, wait for voltage stabilization
    if ((Probe & (1 << REFS1)) != 0) 
    {
      delay(10); // 10 mS for voltage stabilization because Nano has 100nF on Pin AREF
    }
  #endif

  // always do one dummy read of ADC, 112us
  StartADCwait();   // start ADC and wait

  // sample ADC readings
  Value = 0UL;      // reset sampling variable
  Samples = 0;      // number of samples to take

  while (Samples < ADCconfig.Samples) 
  {   // take samples
    StartADCwait(); // start ADC and wait
    Value += ADCW;  // add ADC reading

    #ifdef AUTOSCALE_ADC
      // auto-switch voltage reference for low readings
      if ((Samples == 4) && (ADCconfig.V_Bandgap > 255) && ((uint16_t)Value < 1024) && !(Probe & (1 << REFS1))) 
      {
        Probe |= (1 << REFS1); // select internal bandgap reference
        goto sample;           // re-run sampling
      }
    #endif

    Samples++;    // one more done
  }

  #ifdef AUTOSCALE_ADC
    // convert ADC reading to voltage - single sample: U = ADC reading * V_ref / 1024
    // get voltage of reference used
    if (Probe & (1 << REFS1)) U = ADCconfig.V_Bandgap;  // bandgap reference
    else U = ADCconfig.V_AVCC;                          // Vcc reference
  #else                                 
    U = ADCconfig.V_AVCC;                               // Vcc reference
  #endif

  // convert to voltage
  Value *= U;     // ADC readings * V_ref
  Value /= 1023;  // / 1024 for 10bit ADC

  // de-sample to get average voltage
  Value /= ADCconfig.Samples;
  U = (unsigned int)Value;
  return U;
  //return ((unsigned int)(Value / (1023 * (unsigned long)ADCconfig.Samples)));
}

// ====================================================================================
//  W5msReadADC - W10msReadADC - W20msReadADC
// ====================================================================================
unsigned int W5msReadADC (uint8_t Probe) 
{
  delay(5);
  return (ReadADC(Probe));
}

unsigned int W10msReadADC (uint8_t Probe) 
{
  delay(10);
  return (ReadADC(Probe));
}

unsigned int W20msReadADC (uint8_t Probe) 
{
  delay(20);
  return (ReadADC(Probe));
}

// ====================================================================================
//  READ CAPACITY
// ====================================================================================
// new code by K.-H. Kubbeler
// ReadCapacity tries to find the value of a capacitor by measuring the load time.
// first of all the capacitor is discharged.
// Then a series of up to 500 load pulses with 10ms duration each is done across the R_L (680Ohm)
// resistor.
// After each load pulse the voltage of the capacitor is measured without any load current.
// If voltage reaches a value of more than 300mV and is below 1.3V, the capacity can be
// computed from load time and voltage by a interpolating a build in table.
// If the voltage reaches a value of more than 1.3V with only one load pulse,
// another measurement methode is used:
// The build in 16bit counter can save the counter value at external events.
// One of these events can be the output change of a build in comparator.
// The comparator can compare the voltage of any of the ADC input pins with the voltage
// of the internal reference (1.3V or 1.1V).
// After setting up the comparator and counter properly, the load of capacitor is started 
// with connecting the positive pin with the R_H resistor (470kOhm) to VCC and immediately
// the counter is started. By counting the overflow Events of the 16bit counter and watching
// the counter event flag  the total load time of the capacitor until reaching the internal
// reference voltage can be measured.
// If any of the tries to measure the load time is successful,
// the following variables are set:
// cap.cval = value of the capacitor 
// cap.cval_uncorrected = value of the capacitor uncorrected
// cap.esr = serial resistance of capacitor,  0.01 Ohm units
// cap.cpre = units of cap.cval (-12==pF, -9=nF, -6=uF)
// ca   = Pin number (0-2) of the LowPin
// cb   = Pin number (0-2) of the HighPin
// ====================================================================================
void ReadCapacity(uint8_t HighPin, uint8_t LowPin) 
{
  // check if capacitor and measure the capacity value
  unsigned int tmpint;
  unsigned int adcv[4];
  unsigned int ovcnt16;
  uint8_t HiPinR_L, HiPinR_H;
  uint8_t LoADC;
  uint8_t ii;

  unsigned int vloss;   // lost voltage after load pulse in 0.1% 

  LoADC = pgm_read_byte(&PinADCtab[LowPin]) | TXD_MSK;
  HiPinR_L = pgm_read_byte(&PinRLtab[HighPin]); // R_L mask for HighPin R_L load
  HiPinR_H = HiPinR_L + HiPinR_L;               // double for HighPin R_H load

  if(PartFound == PART_RESISTOR) 
  {
    return; // We have found a resistor already 
  }

  for (ii=0; ii<NumOfDiodes; ii++) 
  {
    if ((diodes[ii].Cathode == LowPin) && (diodes[ii].Anode == HighPin) && (diodes[ii].Voltage < 1500)) 
    {
      return;
    }
  }
  
  cap.esr = 0;                // set ESR of capacitor to zero
  vloss = 0;                  // set lost voltage to zero

  cap.cval = 0;               // set capacity value to zero
  cap.cpre = -12;             // default unit is pF
  DischargePins();            // discharge capacitor

  ADC_PORT = TXD_VAL;         // switch ADC-Port to GND
  PDatSet(0);                 // switch R-Port to GND
  ADC_DDR = LoADC;            // switch Low-Pin to output (GND)
  PDirSet(HiPinR_L);          // switch R_L port for HighPin to output (GND)
  adcv[0] = ReadADC(HighPin); // voltage before any load 

  // ******** should adcv[0] be measured without current???
  adcv[2] = adcv[0];        // preset to prevent compiler warning

  for (ovcnt16=0; ovcnt16<500; ovcnt16++) 
  {
    PDatSet(HiPinR_L);      // R_L to 1 (VCC) 
    PDirSet(HiPinR_L);      // switch Pin to output, across R to GND or VCC
    delay(10);              // wait 10ms

    PDirSet(0);             // switch back to input
    PDatSet(0);             // no Pull up
    delayMicroseconds(500); // wait a little time

    // read voltage without current, is already charged enough?
    adcv[2] = ReadADC(HighPin);

    if (adcv[2] > adcv[0]) 
    {
      adcv[2] -= adcv[0];   // difference to beginning voltage
    } 
    else 
    {
      adcv[2] = 0;          // voltage is lower or same as beginning voltage
    }

    if ((ovcnt16 == 126) && (adcv[2] < 75)) 
    {
      // 300mV can not be reached well-timed 
      break;    // don't try to load any more
    }

    if (adcv[2] > 300) 
    {
      break;    // probably 100mF can be charged well-timed 
    }
  }

  if (adcv[2] < 301) 
  {
    //if (NumOfDiodes != 0) goto measure_with_rh;
    goto not_cap;   // was never charged enough, >100mF or shorted
  }

  // voltage is rised properly and keeps the voltage enough
  if ((ovcnt16 == 0 ) && (adcv[2] > 1300)) 
  {
    goto measure_with_rh;   // Voltage of more than 1300mV is reached in one pulse, too fast loaded
  }

  // Capacity is more than about 50uF

  #ifdef NO_CAP_HOLD_TIME
    ChargePin10ms(HiPinR_H,0);            // switch HighPin with R_H 10ms auf GND, then currentless
    adcv[3] = ReadADC(HighPin) - adcv[0]; // read voltage again, is discharged only a little bit ?

    if (adcv[3] > adcv[0]) 
    {
      adcv[3] -= adcv[0];                 // difference to beginning voltage
    } 
    else 
    {
      adcv[3] = 0;                        // voltage is lower to beginning voltage
    }

    if ((adcv[3] + adcv[3]) < adcv[2]) 
    {
      if (ovcnt16 == 0 )  
      {
        goto measure_with_rh;   // Voltage of more than 1300mV is reached in one pulse, but not hold
      }

      goto not_cap;             // implausible, not yet the half voltage
    }

    cap.cval_uncorrected.dw = ovcnt16 + 1;
    cap.cval_uncorrected.dw *= getRLmultip(adcv[2]);  // get factor to convert time to capacity from table

  #else

    // wait the half the time which was required for loading
    adcv[3] = adcv[2];             // preset to prevent compiler warning

    for (tmpint=0; tmpint<=ovcnt16; tmpint++) 
    {
      delay(5);
      adcv[3] = ReadADC(HighPin);  // read voltage again, is discharged only a little bit ?
    }

    if (adcv[3] > adcv[0]) 
    {
      adcv[3] -= adcv[0];          // difference to beginning voltage
    } 
    else 
    {
      adcv[3] = 0;                 // voltage is lower or same as beginning voltage
    }

    if (adcv[2] > adcv[3]) 
    {
      adcv[3] = adcv[2] - adcv[3]; // build difference to load voltage // lost voltage during load time wait
    } 
    else
    {
      adcv[3] = 0;                 // no lost voltage
    }

    // compute equivalent parallel resistance from voltage drop
    if (adcv[3] > 0) 
    {
      // there is any voltage drop (adcv[3]) !
      // adcv[2] is the loaded voltage.
      vloss = (unsigned long)(adcv[3] * 1000UL) / adcv[2];
    }

    if (adcv[3] > 100)  // more than 100mV lost during load time
    {
      if (ovcnt16 == 0 )  
      {
        goto measure_with_rh; // Voltage of more than 1300mV is reached in one pulse, but not hold
      }

      goto not_cap;           // capacitor does not keep the voltage about 5ms
    }

    cap.cval_uncorrected.dw = ovcnt16 + 1;
    // compute factor with load voltage + lost voltage during the voltage load time
    cap.cval_uncorrected.dw *= getRLmultip(adcv[2]+adcv[3]);  // get factor to convert time to capacity from table
  #endif

  cap.cval = cap.cval_uncorrected.dw; // set result to uncorrected
  cap.cpre = -9;                      // switch units to nF 
  Scale_C_with_vcc();

  // cap.cval for this type is at least 40000nF, so the last digit will be never shown
  cap.cval *= (1000 - REF_C_HI_KORR); // correct with REF_C_HI_KORR with 0.1% resolution, but prevent overflow
  cap.cval /= 100;

  goto checkDiodes;


  // ====================================================================================
  // Measurement of little capacity values
  // ====================================================================================
  measure_with_rh:
  
    // little capacity value, about  < 50 uF
    DischargePins();      // discharge capacitor

    // measure with the R_H (470kOhm) resistor 
    PDatSet(0);   // HiPinR_L
    ADC_DDR = (1<<TP1) | (1<<TP2) | (1<<TP3);  // switch all Pins to output
    ADC_PORT = TXD_VAL;                        // switch all ADC Pins to GND
    PDirSet(HiPinR_H);                         // switch R_H resistor port for HighPin to output (GND)

    // setup Analog Comparator
    ADC_COMP_CONTROL = (1<<ACME);              // enable Analog Comparator Multiplexer
    ACSR =  (1<<ACBG) | (1<<ACI)  | (1<<ACIC); // enable, 1.3V, no Interrupt, Connect to Timer1 
    ADMUX = (1<<REFS0) | HighPin;              // switch Mux to High-Pin
    ADCSRA = (1<<ADIF) | ADC_CLOCK_PRESCALER;  // disable ADC
    delayMicroseconds(200);                    // wait for bandgap to start up

    // setup Counter1
    ovcnt16 = 0;
    TCCR1A = 0;     // set Counter1 to normal Mode
    TCNT1 = 0;      // set Counter to 0
    TI1_INT_FLAGS = (1<<ICF1) | (1<<OCF1B) | (1<<OCF1A) | (1<<TOV1);  // clear interrupt flags

    PDatSet(HiPinR_H);          // switch R_H resistor port for HighPin to VCC

    if(PartFound == PART_FET) 
    {
      // charge capacitor with R_H resistor
      TCCR1B = (1<<CS10); //Start counter 1MHz or 8MHz
      ADC_DDR = (((1<<TP1) | (1<<TP2) | (1<<TP3) | TXD_MSK) & ~(1<<HighPin)); // release only HighPin ADC port
    } 
    else 
    {
      TCCR1B =  (1<<CS10); // start counter 1MHz or 8MHz
      ADC_DDR = LoADC;     // stay LoADC Pin switched to GND, charge capacitor with R_H slowly
    }

    //******************************************************

    while(1)                 // Wait, until  Input Capture is set
    {
      ii = TI1_INT_FLAGS;    // read Timer flags

      if (ii & (1<<ICF1))  
      {
        break;
      }

      if((ii & (1<<TOV1)))          // counter overflow, 65.536 ms @ 1MHz, 8.192ms @ 8MHz
      {
        TI1_INT_FLAGS = (1<<TOV1);  // Reset OV Flag
        ovcnt16++;

        if(ovcnt16 == (F_CPU/5000))
        {
          break;                    // Timeout for Charging, above 12 s
        }
      }
    }

    TCCR1B = (0<<ICNC1) | (0<<ICES1) | (0<<CS10);  // stop counter
    TI1_INT_FLAGS = (1<<ICF1);                     // Reset Input Capture
    tmpint = ICR1;                                 // get previous Input Capture Counter flag

    // check actual counter, if an additional overflow must be added
    if((TCNT1 > tmpint) && (ii & (1<<TOV1))) 
    {
      // this OV was not counted, but was before the Input Capture
      TI1_INT_FLAGS = (1<<TOV1);  // Reset OV Flag
      ovcnt16++;
    }

    //------------------------------------------------------------
    ADCSRA = (1<<ADEN) | (1<<ADIF) | ADC_CLOCK_PRESCALER;  // enable ADC
    PDirSet(0);                                            // switch R_H resistor port for input
    PDatSet(0);                                            // switch R_H resistor port pull up for HighPin off
    adcv[2] = ReadADC(HighPin);                            // get loaded voltage
    load_diff = adcv[2] + REF_C_LOW_KORR - ref_mv;         // build difference of capacitor voltage to Reference Voltage
    //------------------------------------------------------------

    if (ovcnt16 >= (F_CPU/10000)) 
    {
      goto not_cap; // no normal end
    }

    //cap.cval_uncorrected = CombineII2Long(ovcnt16, tmpint);
    cap.cval_uncorrected.w[1] = ovcnt16;
    cap.cval_uncorrected.w[0] = tmpint;

    cap.cpre = -12;                   // cap.cval unit is pF 
    if (ovcnt16 > 65) 
    {
      cap.cval_uncorrected.dw /= 100; // switch to next unit
      cap.cpre += 2;                  // set unit, prevent overflow
    }

    cap.cval_uncorrected.dw *= RHmultip;        // 708
    cap.cval_uncorrected.dw /= (F_CPU / 10000); // divide by 100 (@ 1MHz clock), 800 (@ 8MHz clock)
    cap.cval = cap.cval_uncorrected.dw;         // set the corrected cap.cval
    Scale_C_with_vcc();

    if (cap.cpre == -12) 
    {
      #if COMP_SLEW1 > COMP_SLEW2
        if (cap.cval < COMP_SLEW1)
        {
          cap.cval += (COMP_SLEW1 / (cap.cval+COMP_SLEW2 )); // add slew rate dependent offset
        }
      #endif

      if (cap.cval > C_NULL) 
      {
        cap.cval -= C_NULL;   // subtract constant offset (pF)
      } 
      else 
      {
        cap.cval = 0;         // unsigned long may not reach negativ value
      }
    }

    PDirSet(HiPinR_L);  // switch R_L for High-Pin to GND

    if(cap.cval < 25)   // cap.cval can only be so little in pF unit, cap.cpre must not be testet!
    {
      goto not_cap;     // capacity to low, < 50pF @1MHz (25pF @8MHz)
    }

  // end low capacity 


  // ====================================================================================
  //  CHECK DIODES
  // ====================================================================================
  checkDiodes:

    if((NumOfDiodes > 0)  && (PartFound != PART_FET)) 
    {
      // nearly shure, that there is one or more diodes in reverse direction,
      // which would be wrongly detected as capacitor 
    } 
    else 
    {
      PartFound = PART_CAPACITOR;   // capacitor is found

      if ((cap.cpre > cap.cpre_max) || ((cap.cpre == cap.cpre_max) && (cap.cval > cap.cval_max))) 
      {
        // we have found a greater one
        cap.cval_max = cap.cval;
        cap.cpre_max = cap.cpre;
        
        cap.v_loss = vloss; // lost voltage in 0.01%

        cap.ca = LowPin;    // save LowPin
        cap.cb = HighPin;   // save HighPin
      }
    }

  not_cap:

    // discharge capacitor again
    // DischargePins();   // discharge capacitors
    // ready
    // switch all ports to input
    ADC_DDR =  TXD_MSK;   // switch all ADC ports to input
    ADC_PORT = TXD_VAL;   // switch all ADC outputs to GND, no pull up
    PDirSet(0);           // switch all resistor ports to input
    PDatSet(0);           // switch all resistor outputs to GND, no pull up

    return;
}  // end ReadCapacity()


// ====================================================================================
//  GET RL MULTIP
// ====================================================================================
// interpolate table RLtab corresponding to voltage cvolt
// ====================================================================================
unsigned int getRLmultip(unsigned int cvolt) 
{
  #define RL_Tab_Abstand 25     // displacement of table 25mV
  #define RL_Tab_Beginn 300     // begin of table ist 300mV
  #define RL_Tab_Length 1100    // length of table is 1400-300

  unsigned int uvolt;
  unsigned int y1, y2;
  uint8_t tabind;
  uint8_t tabres;

  if (cvolt >= RL_Tab_Beginn) 
  {
    uvolt = cvolt - RL_Tab_Beginn;
  } 
  else 
  {
    uvolt = 0;      // limit to begin of table
  }

  tabind = uvolt / RL_Tab_Abstand;
  tabres = uvolt % RL_Tab_Abstand;
  tabres = RL_Tab_Abstand - tabres;

  if (tabind > (RL_Tab_Length/RL_Tab_Abstand)) 
  {
    tabind = (RL_Tab_Length/RL_Tab_Abstand);  // limit to end of table
  }

  y1 = pgm_read_word(&RLtab[tabind]);
  y2 = pgm_read_word(&RLtab[tabind+1]);
  return (((y1 - y2) * tabres + (RL_Tab_Abstand/2)) / RL_Tab_Abstand + y2); // interpolate table
}

// ====================================================================================
//  SCALE C WITH VCC
// ====================================================================================
void Scale_C_with_vcc(void) 
{

  while (cap.cval > 100000) 
  {
    cap.cval /= 10;
    cap.cpre ++;                  // prevent overflow
  }

  cap.cval *= ADCconfig.V_AVCC;   // scale with measured voltage
  cap.cval /= V_VCC;              // Factors are computed for V_VCC
}


// ====================================================================================
//  READ INDUCTANCE
// ====================================================================================
// new code by K.-H. Kubbeler
// The 680 Ohm resistor (R_L_VAL) at the Lowpin will be used as current sensor
// The current with a coil will with (1 - e**(-t*R/L)), where R is
// the sum of Pin_RM , R_L_VAL , Resistance of coil and Pin_RP.
// L in the inductance of the coil.
// ====================================================================================
void ReadInductance(void) 
{
  // check if inductor and measure the inductance value
  unsigned int tmpint;
  unsigned int v_max;
  unsigned int total_r;        // total resistance of current loop
  unsigned int mess_r;         // value of resistor used for current measurement
  unsigned long inductance[4]; // four inductance values for different measurements

  union t_combi
  {
    unsigned long dw;         // time_constant
    uint16_t w[2];
  } timeconstant;

  uint16_t per_ref1,per_ref2; // percentage
  uint8_t LoPinR_L;           // Mask for switching R_L resistor of low pin
  uint8_t HiADC;              // Mask for switching the high pin direct to VCC
  uint8_t ii;
  uint8_t count;              // counter for the different measurements

  //uint8_t found;            // variable used for searching resistors 
  #define found 0

  uint8_t cnt_diff;           // resistance dependent offset
  uint8_t LowPin;             // number of pin with low voltage
  uint8_t HighPin;            // number of pin with high voltage 
  int8_t  v_corr;             // correction of comparator voltage
  uint8_t nr_pol1;            // number of successfull inductance measurement with polarity 1
  uint8_t nr_pol2;            // number of successfull inductance measurement with polarity 2

  if(PartFound != PART_RESISTOR) 
  {
    return; // We have found no resistor
  }

  if (ResistorsFound != 1) 
  {
    return; // do not search for inductance, more than 1 resistor
  }

  if (resis[found].rx > 21000) return;
  // we can check for Inductance, if resistance is below 2100 Ohm

  for (count=0; count<4; count++) // Try four times (different direction and with delayed counter start)
  {
    if (count < 2) 
    {
      // first and second pass, direction 1
      LowPin = resis[found].ra;
      HighPin = resis[found].rb;
    } 
    else 
    {
      // third and fourth pass, direction 2
      LowPin = resis[found].rb;
      HighPin = resis[found].ra;
    }

    HiADC = pgm_read_byte(&PinADCtab[HighPin]);
    LoPinR_L = pgm_read_byte(&PinRLtab[LowPin]);  // R_L mask for HighPin R_L load

    //======================================================================================
    // Measurement of Inductance values
    PDatSet(0);           // switch R port to GND
    ADC_PORT = TXD_VAL;   // switch ADC-Port to GND

    if ((resis[found].rx < 240) && ((count & 0x01) == 0)) 
    {
      // we can use PinR_L for measurement
      mess_r = RR680MI - R_L_VAL;               // use only pin output resistance
      ADC_DDR = HiADC | (1<<LowPin) | TXD_MSK;  // switch HiADC and Low Pin to GND, 
    } 
    else 
    {
      PDirSet(LoPinR_L);          // switch R_L resistor port for LowPin to output (GND)
      ADC_DDR = HiADC | TXD_MSK;  // switch HiADC Pin to GND 
      mess_r = RR680MI;           // use 680 Ohm and PinR_L for current measurement
    }

    // Look, if we can detect any current
    for (ii=0; ii<20; ii++)                     // wait for current near to zero
    {
      v_max = W10msReadADC(LowPin);
      total_r =  ReadADC(HighPin);
      if ((v_max < 2) && (total_r < 2)) break;  // low current detected
    }

    // setup Analog Comparator
    ADC_COMP_CONTROL = (1<<ACME);               // enable Analog Comparator Multiplexer
    ACSR =  (1<<ACBG) | (1<<ACI)  | (1<<ACIC);  // enable, 1.3V, no Interrupt, Connect to Timer1 
    ADMUX = (1<<REFS0) | LowPin;                // switch Mux to Low-Pin
    ADCSRA = (1<<ADIF) | ADC_CLOCK_PRESCALER;   // disable ADC
   
    // setup Counter1
    timeconstant.w[1] = 0;  // set ov counter to 0
    TCCR1A = 0;             // set Counter1 to normal Mode
    TCNT1 = 0;              // set Counter to 0
    TI1_INT_FLAGS = (1<<ICF1) | (1<<OCF1B) | (1<<OCF1A) | (1<<TOV1);  // reset TIFR or TIFR1
    HiADC |= TXD_VAL;
    delayMicroseconds(200); // wait for bandgap to start up

    if ((count & 0x01) == 0) 
    {
      // first start counter, then start current
      TCCR1B =  (1<<ICNC1) | (0<<ICES1) | (1<<CS10); // start counter 1MHz or 8MHz
      ADC_PORT = HiADC;                              // switch ADC-Port to VCC
    } 
    else 
    {
      // first start current, then start counter with delay
      // parasitic capacity of coil can cause high current at the beginning
      ADC_PORT = HiADC;                              // switch ADC-Port to VCC
      delayMicroseconds(3);                          // ignore current peak from capacity
      TI1_INT_FLAGS = (1<<ICF1);                     // Reset Input Capture
      TCCR1B =  (1<<ICNC1) | (0<<ICES1) | (1<<CS10); // start counter 1MHz or 8MHz
    }
      
    while(1)  // Wait, until  Input Capture is set
    {
      ii = TI1_INT_FLAGS;   // read Timer flags

      if (ii & (1<<ICF1))  
      {
        break;
      }

      if((ii & (1<<TOV1)))  // counter overflow, 65.536 ms @ 1MHz, 8.192ms @ 8MHz
      { 
        TI1_INT_FLAGS = (1<<TOV1);  // Reset OV Flag

        timeconstant.w[1]++;        // count one OV

        if(timeconstant.w[1] == (F_CPU/100000UL)) 
        {
          break;                    // Timeout for Charging, above 0.13 s
        }
      }
    }

    TCCR1B = (0<<ICNC1) | (0<<ICES1) | (0<<CS10);   // stop counter
    TI1_INT_FLAGS = (1<<ICF1);                      // Reset Input Capture
    timeconstant.w[0] = ICR1;                       // get previous Input Capture Counter flag

    // check actual counter, if an additional overflow must be added
    if((TCNT1 > timeconstant.w[0]) && (ii & (1<<TOV1))) 
    {
      // this OV was not counted, but was before the Input Capture
      TI1_INT_FLAGS = (1<<TOV1);  // Reset OV Flag
      timeconstant.w[1]++;        // count one additional OV
    }

    ADC_PORT = TXD_VAL;                                   // switch ADC-Port to GND
    ADCSRA = (1<<ADEN) | (1<<ADIF) | ADC_CLOCK_PRESCALER; // enable ADC

    for (ii=0; ii<20; ii++)       // wait for current near to zero
    {
      v_max = W10msReadADC(LowPin);
      total_r =  ReadADC(HighPin);

      if ((v_max < 2) && (total_r < 2)) break;   // low current detected
    }

    #define CNT_ZERO_42 6
    #define CNT_ZERO_720 7

    total_r = (mess_r + resis[found].rx + RRpinMI);

    cnt_diff = total_r / ((14000UL * 8) / (F_CPU/1000000UL));
    
    // Voltage of comparator in % of v_max
    tmpint = (ref_mv + REF_C_LOW_KORR);

    if (mess_r < R_L_VAL)
    {
      // ------------------------- measurement with 680 Ohm
      cnt_diff = CNT_ZERO_42;

      if (timeconstant.dw < 225) 
      {
        v_corr = (timeconstant.w[0] / 5) - 20;
      }
      else 
      {
        v_corr = 25;
      }

      tmpint += (REF_L_LOW_KORR - v_corr);
    } 
    else 
    {
      // ------------------------ measurement without 680 Ohm resistor
      cnt_diff += CNT_ZERO_720;
      tmpint += REF_L_HI_KORR;
    }

    if (timeconstant.dw > cnt_diff) timeconstant.dw -= cnt_diff;
    else timeconstant.dw = 0;
       
    if ((count&0x01) == 1) 
    {
      // second pass with delayed counter start
      timeconstant.dw += (3 * (F_CPU/1000000UL))+10;
    }

    if (timeconstant.w[1] >= (F_CPU/100000UL)) timeconstant.dw = 0;  // no transition found

    if (timeconstant.dw > 10) 
    {
      timeconstant.dw -= 1;
    }

    // compute the maximum Voltage v_max with the Resistor of the coil
    v_max = ((unsigned long)mess_r * (unsigned long)ADCconfig.V_AVCC) / total_r;
    per_ref1 = ((unsigned long)tmpint * 1000) / v_max;
    per_ref2 = get_log(per_ref1);       // -log(1 - per_ref1/1000)

    //*********************************************************
    // lx in 0.01mH units, L = Tau * R
    per_ref1 = ((per_ref2 * (F_CPU / 1000000UL)) + 5) / 10;
    inductance[count] = (timeconstant.dw * total_r ) / per_ref1;

    if (((count&0x01) == 0) && (timeconstant.dw > ((F_CPU / 1000000UL) + 3))) 
    {
      // transition is found, measurement with delayed counter start is not necessary
      inductance[count+1] = inductance[count];  // set delayed measurement to same value
      count++;          // skip the delayed measurement
    }

  }  // end for count

  ADC_PORT = TXD_VAL;   // switch ADC Port to GND
  delay(20);

  #if 0

    if (inductance[1] > inductance[0]) 
    {
      resis[found].lx = inductance[1];  // use value found with delayed counter start
    } 
    else 
    {
      resis[found].lx = inductance[0];
    }

    if (inductance[3] > inductance[2]) inductance[2] = inductance[3];   // other polarity, delayed start

    if (inductance[2] < resis[found].lx) resis[found].lx = inductance[2]; // use the other polarity

  #else

    nr_pol1 = 0;
    if (inductance[1] > inductance[0]) { nr_pol1 = 1; } 

    nr_pol2 = 2;
    if (inductance[3] > inductance[2]) { nr_pol2 = 3; } 

    if (inductance[nr_pol2] < inductance[nr_pol1]) nr_pol1 = nr_pol2;

    resis[found].lx = inductance[nr_pol1];
    resis[found].lpre = -5;               // 10 uH units

    if (((nr_pol1 & 1) == 1) || (resis[found].rx >= 240)) 
    {
      // with 680 Ohm resistor total_r is more than 7460
      resis[found].lpre = -4;         // 100 uH units
      resis[found].lx = (resis[found].lx + 5) / 10;
    } 

  #endif

  // ---------------------- switch all ports to input
  ADC_DDR =  TXD_MSK;   // switch all ADC ports to input
  PDirSet(0);           // switch all resistor ports to input

  return;
}  // end ReadInductance()


// ====================================================================================
//  GET LOG
// ====================================================================================
// Interpolate the table LogTab[] with the function -1000*log(1 - (permil/1000))
// ====================================================================================
uint16_t get_log(uint16_t permil) 
{
  #define Log_Tab_Distance 20              // displacement of table is 20 mil

  uint16_t y1, y2;    // table values
  uint16_t result;    // result of interpolation
  uint8_t tabind;     // index to table value
  uint8_t tabres;     // distance to lower table value, fraction of Log_Tab_Distance

  tabind = permil / Log_Tab_Distance;    // index to table
  tabres = permil % Log_Tab_Distance;    // fraction of table distance

  // interpolate the table of factors
  y1 = pgm_read_word(&LogTab[tabind]);   // get the lower table value
  y2 = pgm_read_word(&LogTab[tabind+1]); // get the higher table value
  result = ((y2 - y1) * tabres ) / Log_Tab_Distance + y1; // interpolate

  return(result);
 }


// ################################################################################
// ################################################################################
// ################################################################################
//  WAIT TIMES
// ################################################################################
// ################################################################################
// ################################################################################

// --------------------------------------------------------------------------------
//  A delayMicroseconds(20) is used to switch off the load current for big caps
//   when time is too short, voltage is down before SH of ADC
//   when time is too long, capacitor will be overloaded.
//   That will cause too high voltage without current.
// --------------------------------------------------------------------------------

// --------------------------------------------------------------------------------
//  A short delay of 62.5 ns is used in functions GetESR and GetVLOSS
// --------------------------------------------------------------------------------
#define wait62p5ns() __asm__ __volatile__ ("nop") // one cycle @ 16MHz = 62.5 nS


// ====================================================================================
//  GET ESR - measure the ESR value
//  This function is called from capacitors and low resistors (< 100 ohm)
// ====================================================================================
#define MAX_CNT 255

uint16_t GetESR(uint8_t hipin, uint8_t lopin) 
{
  unsigned int adcv[4];     // array for 4 ADC readings
  unsigned long sumvolt[4]; // array for 3 sums of ADC readings
  uint16_t esrvalue;
  uint8_t HiPinR_L;         // used to switch 680 Ohm to HighPin
  uint8_t HiADC;            // used to switch Highpin directly to GND or VCC
  uint8_t LoPinR_L;         // used to switch 680 Ohm to LowPin
  uint8_t LoADC;            // used to switch Lowpin directly to GND or VCC
  uint8_t ii;               // tempory values
  uint8_t SelectLowPin;
  uint8_t SelectHighPin;

  LoADC = pgm_read_byte(&PinADCtab[lopin]) | TXD_MSK;
  HiADC = pgm_read_byte(&PinADCtab[hipin]) | TXD_MSK;
  LoPinR_L = pgm_read_byte(&PinRLtab[lopin]);        // R_L mask for LowPin R_L load
  HiPinR_L = pgm_read_byte(&PinRLtab[hipin]);        // R_L mask for HighPin R_L load

  SelectLowPin = (lopin | (1<<REFS1) | (1<<REFS0));  // switch ADC to LowPin, Internal Ref. 
  SelectHighPin = (hipin | (1<<REFS1) | (1<<REFS0)); // switch ADC to HighPin, Internal Ref. 

  // Measurement of ESR of capacitors AC Mode
  sumvolt[0] = 1;             // set sum of LowPin voltage to 1 to prevent divide by zero
  sumvolt[2] = 1;             // clear sum of HighPin voltage with current

  // offset is about (x*10*200)/34000 in 0.01 Ohm units
  
  sumvolt[1] = 0;             // clear sum of HighPin voltage without current
  sumvolt[3] = 0;             // clear sum of HighPin voltage without current
  DischargePins();            // discharge capacitor
  ADC_PORT = TXD_VAL;         // switch ADC-Port to GND
  ADMUX = SelectLowPin;       // set Mux input and Voltage Reference to internal 1.1V

  delay(10); // 10 mS for voltage stabilization because Nano has 100nF on Pin AREF

  // start voltage must be negative
  ADC_DDR = HiADC;            // switch High Pin to GND
  PDatSet(LoPinR_L);          // switch R-Port to VCC
  PDirSet(LoPinR_L);          // switch R_L port for HighPin to output (VCC)
  delayMicroseconds(10);
  delayMicroseconds(2);
  PDirSet(0);                 // switch off current
  PDatSet(0);
  StartADCwait();             // set ADCSRA Interrupt Mode, sleep

  // Measurement frequency is given by sum of ADC-Reads < 680 Hz for normal ADC speed.
  // For fast ADC mode the frequency is below 2720 Hz (used for capacity value below 3.6 uF).
  // ADC Sample and Hold (SH) is done 1.5 ADC clock number after real start of conversion.
  // Real ADC-conversion is started with the next ADC-Clock (125kHz) after setting the ADSC bit.

  for(ii=0; ii<MAX_CNT; ii++) 
  {
    ADC_DDR = LoADC;          // switch Low-Pin to output (GND)
    PDatSet(LoPinR_L);        // switch R-Port to VCC
    PDirSet(LoPinR_L);        // switch R_L port for LowPin to output (VCC)
    ADMUX = SelectLowPin;
    StartADCwait();           // set ADCSRA Interrupt Mode, sleep
    StartADCwait();           // set ADCSRA Interrupt Mode, sleep
    adcv[0] = ADCW;           // Voltage LowPin with current
    ADMUX = SelectHighPin;

    StartADCwait();     // ADCSRA = (1<<ADEN) | (1<<ADIF) | (1<<ADIE) | ADC_CLOCK_PRESCALER;  
    ADCSRA = (1<<ADSC) | (1<<ADEN) | (1<<ADIF) | ADC_CLOCK_PRESCALER; // enable ADC and start with ADSC
    delayMicroseconds(4);
    PDatSet(HiPinR_L);        // switch R-Port to VCC
    PDirSet(HiPinR_L);        // switch R_L port for HighPin to output (VCC)
    delayMicroseconds(20);    // wait predefined time

    PDirSet(0);               // switch current off,  SH is 1.5 ADC clock behind real start
    PDatSet(0);
    while (ADCSRA&(1<<ADSC)); // wait for conversion finished
    adcv[1] = ADCW;           // Voltage HighPin with current

    // ******** Reverse direction, connect High side with GND ********
    ADC_DDR = HiADC;          // switch High Pin to GND
    PDatSet(HiPinR_L);        // switch R-Port to VCC
    PDirSet(HiPinR_L);        // switch R_L port for HighPin to output (VCC)
    
    //delayMicroseconds(10);  // wait a little time
    wait62p5ns();
    
    ADMUX = SelectHighPin;
    StartADCwait();           // set ADCSRA Interrupt Mode, sleep
    StartADCwait();           // set ADCSRA Interrupt Mode, sleep
    adcv[2] = ADCW;           // Voltage HighPin with current
    ADMUX = SelectLowPin;

    StartADCwait();           // set ADCSRA Interrupt Mode, sleep
    ADCSRA = (1<<ADSC) | (1<<ADEN) | (1<<ADIF) | ADC_CLOCK_PRESCALER; // enable ADC and start with ADSC
    delayMicroseconds(4);
    PDatSet(LoPinR_L);
    PDirSet(LoPinR_L);        // switch LowPin with 680 Ohm to VCC
    delayMicroseconds(20);    // wait predefined time

    PDirSet(0);               // switch current off
    PDatSet(0);

    while (ADCSRA&(1<<ADSC)); // wait for conversion finished
    adcv[3] = ADCW;           // Voltage LowPin with current

    sumvolt[0] += adcv[0];    // add sum of both LowPin voltages with current
    sumvolt[1] += adcv[1];    // add  HighPin voltages with current
    sumvolt[2] += adcv[2];    // add  LowPin voltages with current
    sumvolt[3] += adcv[3];    // add  HighPin voltages with current

  } // end for

  sumvolt[0] += sumvolt[2];

  if ((sumvolt[1] + sumvolt[3]) > sumvolt[0]) 
  {
    sumvolt[2] = (sumvolt[1] + sumvolt[3]) - sumvolt[0];  // difference HighPin - LowPin Voltage with current
  } 
  else 
  {
    sumvolt[2] = 0;
  }

  esrvalue = (sumvolt[2] * 10 * (unsigned long)RRpinMI) / (sumvolt[0] + sumvolt[2]);
  esrvalue += esrvalue / 14;  // esrvalue + 7%

  return (esrvalue);
}


// ====================================================================================
//  GET VLOSS
// ====================================================================================
// new code by K.-H. Kubbeler
// ca   = Pin number (0-2) of the LowPin
// cb   = Pin number (0-2) of the HighPin
//=====================================================================================
void GetVloss() 
{
  // measure voltage drop after load pulse
  unsigned int tmpint;
  unsigned int adcv[4];

  union t_combi
  {
    unsigned long dw;    // capacity value  in 100nF units
    uint16_t w[2];
  } lval;

  uint8_t ii;
  uint8_t HiPinR_L;
  uint8_t LoADC;

  if (cap.v_loss > 0) return; // Voltage loss is already known

  LoADC = pgm_read_byte(&PinADCtab[cap.ca]) | TXD_MSK;
  HiPinR_L = pgm_read_byte(&PinRLtab[cap.cb]);    // R_L mask for HighPin R_L load

  DischargePins();            // discharge capacitor

  ADC_PORT = TXD_VAL;         // switch ADC-Port to GND
  PDatSet(0);                 // switch R-Port to GND
  ADC_DDR = LoADC;            // switch Low-Pin to output (GND)
  PDirSet(HiPinR_L);          // switch R_L port for HighPin to output (GND)
  adcv[0] = ReadADC(cap.cb);  // voltage before any load 

  // ******** should adcv[0] be measured without current???
  if (cap.cpre_max > -9) return; // too much capacity

  lval.dw = cap.cval_max;

  for (ii=cap.cpre_max+12; ii<4; ii++)     // was < 5
  {
    lval.dw = (lval.dw + 5) / 10;
  }

  if ((lval.dw == 0) || (lval.dw > 5000))  // was > 500
  {
    return;  // capacity more than 50uF, Voltage loss is already measured
  }

  PDatSet(HiPinR_L);   // R_L to 1 (VCC) 
  PDirSet(HiPinR_L);   // switch Pin to output, across R to GND or VCC

  for (tmpint=0; tmpint<lval.w[0]; tmpint+=2) 
  {
    delayMicroseconds(5);   
  }

  PDirSet(0);                 // switch back to input
  PDatSet(0);                 // no Pull up

  //delayMicroseconds(10);    // wait a little time
  wait62p5ns();

  // read voltage without current
  ADCconfig.Samples = 5;      // set ADC to only 5 samples
  adcv[2] = ReadADC(cap.cb);
  if (adcv[2] > adcv[0]) 
  {
    adcv[2] -= adcv[0];       // difference to beginning voltage
  } 
  else 
  {
    adcv[2] = 0;              // voltage is lower or same as beginning voltage
  }

  // wait 2x the time which was required for loading
  for (tmpint=0; tmpint<lval.w[0]; tmpint++) 
  {
    delayMicroseconds(5);
  }

  adcv[3] = ReadADC(cap.cb);      // read voltage again, is discharged only a little bit ?
  ADCconfig.Samples = N_MEASURES; // set ADC back to configured No. of samples
  wait62p5ns();

  if (adcv[3] > adcv[0]) 
  {
    adcv[3] -= adcv[0];          // difference to beginning voltage
  } 
  else 
  {
    adcv[3] = 0;                 // voltage is lower or same as beginning voltage
  }

  if (adcv[2] > adcv[3])         // build difference to load voltage
  {                      
    adcv[1] = adcv[2] - adcv[3]; // lost voltage during load time wait
  } 
  else 
  {
    adcv[1] = 0;                 // no lost voltage
  }

  // compute voltage drop as part from loaded voltage
  if (adcv[1] > 0) 
  {
    // there is any voltage drop (adcv[1]) !
    // adcv[2] is the loaded voltage.
    cap.v_loss = (unsigned long)(adcv[1] * 500UL) / adcv[2];
  }

  DischargePins();      // discharge capacitor again

  // switch all ports to input
  ADC_DDR =  TXD_MSK;   // switch all ADC ports to input
  ADC_PORT = TXD_VAL;   // switch all ADC outputs to GND, no pull up
  PDirSet(0);           // switch all resistor ports to input
  PDatSet(0);           // switch all resistor outputs to GND, no pull up

  return;
}


// ====================================================================================
//  CALIBRATE VR
// ====================================================================================
void Calibrate_VR(void) 
{
  // get reference voltage, calibrate VCC with external 2.5V and
  // get the port output resistance

  ADCconfig.V_AVCC = V_VCC;          // set initial VCC Voltage
  ADCconfig.Samples = 190;           // set number of ADC reads near to maximum

  ADC_PORT = TXD_VAL;                // switch to 0V
  ADC_DDR = (1<<TPREF) | TXD_MSK;    // switch pin with 2.5V reference to GND
  delay(1);
  ADC_DDR =  TXD_MSK;                // switch pin with reference back to input
  trans.vBE[1] = W5msReadADC(TPREF); // read voltage of 2.5V precision reference

  if ((trans.vBE[1] > 2250) && (trans.vBE[1] < 2750)) 
  {
    // precision voltage reference connected, update V_AVCC
    WithReference = 1;
    ADCconfig.V_AVCC = (unsigned long)((unsigned long)ADCconfig.V_AVCC * 2495) / trans.vBE[1];
  }

  #ifdef WITH_AUTO_REF
    (void) ReadADC(MUX_INT_REF);       // read reference voltage 
    ref_mv = W5msReadADC(MUX_INT_REF); // read reference voltage 
    RefVoltage();                      // compute RHmultip = f(reference voltage)
  #else
    ref_mv = DEFAULT_BAND_GAP;         // set to default Reference Voltage
  #endif
 
  ADCconfig.V_Bandgap = ADC_internal_reference; // set internal reference voltage for ADC

  ADCconfig.Samples = N_MEASURES;      // set to configured number of ADC samples
} 


// ====================================================================================
//  ShowPinNumber
// ====================================================================================
void ShowPinNumber(unsigned char temp) 
{
  Serial.write(temp + '1'); // from binary 0 we send ASCII 1
}


// ====================================================================================
//  DISPLAY VALUE - display of values and units
// ==================================================================================== 
//   display value and unit
//    - max. 4 digits excluding "." and unit
//   requires:
//    - value
//    - exponent of factor related to base unit (value * 10^x)
//      e.g: p = 10^-12 -> -12
//    - unit character (0 = none)
//      digits = 2, 3 or 4
// ==================================================================================== 
void DisplayValue(unsigned long Value, int8_t Exponent, String Unit, unsigned char digits)
{
  char OutBuffer[15];
  unsigned int      Limit;
  unsigned char     Prefix;   // prefix character
  uint8_t           Offset;   // exponent of offset to next 10^3 step
  uint8_t           Index;    // index ID
  uint8_t           Length;   // string length

  Limit = 100;                    // scale value down to 2 digits
  if (digits == 3) Limit = 1000;  // scale value down to 3 digits
  if (digits == 4) Limit = 10000; // scale value down to 4 digits

  while (Value >= Limit)
  {
    Value += 5;             // for automatic rounding
    Value = Value / 10;     // scale down by 10^1
    Exponent++;             // increase exponent by 1
  }

  // ---------------------------------- determine prefix
  Length = Exponent + 12;
  if ((int8_t)Length <  0) Length = 0;  // Limit to minimum prefix
  if (Length > 18) Length = 18;         // Limit to maximum prefix
  Index = Length / 3;
  Offset = Length % 3;

  if (Offset > 0)
  {
    Index++;                         // adjust index for exponent offset, take next prefix
    Offset = 3 - Offset;             // reverse value (1 or 2)
  }

  Prefix = pgm_read_byte((uint8_t *)(&PrefixTab[Index]));   // look up prefix in table

  // -------------------------------- display value
  // convert value into string
  utoa((unsigned int)Value, OutBuffer, 10);
  Length = strlen(OutBuffer);

  Exponent = Length - Offset;         // calculate position of dot

  if (Exponent <= 0)                  // we have to prepend "0."
  {
    // 0: factor 10 / -1: factor 100
    Serial.print("0");
    Serial.print(".");
    if (Exponent < 0) Serial.print("0"); // extra 0 for factor 100
  }

  if (Offset == 0) Exponent = -1;     // disable dot if not needed

  // display value and add dot if requested
  Index = 0;
  while (Index < Length)            // loop through string
  {
    Serial.write(OutBuffer[Index]);    // display char
    Index++;                        // next one
    if (Index == Exponent) 
    {
      Serial.print(".");            // display dot
    }
  }

  // display prefix and unit
  if (Prefix != 0) Serial.write(Prefix);
  Serial.print(Unit);
}

// ====================================================================================
//  PIN LAYOUT - show the Pin Layout of the device
// ====================================================================================  
void PinLayout(char pin1, char pin2, char pin3) 
{
  // --------------------------- pin 123 = EBC or SGD or CGA
  Serial.print(" 123=");
  for (ii=0; ii<3; ii++) 
  {
    if (ii == trans.e) Serial.write(pin1);  // Output Character in right order
    if (ii == trans.b) Serial.write(pin2);
    if (ii == trans.c) Serial.write(pin3);
  }
}
